from __future__ import print_function
from petram.debug import handle_allow_python_function_coefficient
from petram.phys.vtable import VtableElement, Vtable
from petram.helper.variables import Variable, eval_code
from petram.phys.vtable import VtableElement, Vtable, Vtable_mixin
from petram.helper.variables import NativeCoefficientGenBase, CoefficientVariable
from petram.mfem_config import use_parallel

import numpy as np
from os.path import dirname, basename, isfile, join
import warnings
import glob
import types
import numbers
from abc import abstractmethod

import petram
from petram.model import Model, Bdry, Domain
from petram.namespace_mixin import NS_mixin

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('Phys')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem


# not that PyCoefficient return only real number array


class PhysConstant(mfem.ConstantCoefficient):
    def __init__(self, value):
        self.value = value
        mfem.ConstantCoefficient.__init__(self, value)

    def __repr__(self):
        return self.__class__.__name__ + "(" + str(self.value) + ")"


class PhysVectorConstant(mfem.VectorConstantCoefficient):
    def __init__(self, value):
        self._value = mfem.Vector(value)
        self.value = self._value.GetDataArray()
        mfem.VectorConstantCoefficient.__init__(self, value)

    def __repr__(self):
        return self.__class__.__name__ + "(" + str(self._value) + ")"


class PhysMatrixConstant(mfem.MatrixConstantCoefficient):
    def __init__(self, value):
        v = mfem.Vector(np.transpose(value).flatten())
        m = mfem.DenseMatrix(v.GetData(), value.shape[0], value.shape[1])
        self._value = (v, m)
        self.value = m.GetDataArray()
        mfem.MatrixConstantCoefficient.__init__(self, m)

    def __repr__(self):
        return self.__class__.__name__ + "(" + str(self._value) + ")"


def try_eval(exprs, l, g):
    '''
    first evaulate as it is
    if the result is list.. return
    if not evaluate w/o knowing any Variables
    '''
    try:
        value = eval(exprs, l, g)
        if isinstance(value, list):
            return True, [value]
        ll = [x for x in l if not isinstance(x, Variable)]
        gg = [x for x in g if not isinstance(x, Variable)]
        value = eval(exprs, ll, gg)
        return True, [value]
    except BaseException:
        return False, exprs


class Coefficient_Evaluator(object):
    def __init__(self, exprs, ind_vars, l, g, real=True):
        '''
        this is complicated....
           elemental (or array) form
            [1,a,3]  (a is namespace variable) is given as [[1, a (valued), 3]]

           single box
              1+1j   is given as '(1+1j)' (string)

           matrix
              given as string like '[0, 1, 2, 3, 4]'
              if variable is used it is become string element '['=variable', 1, 2, 3, 4]'
              if =Varialbe in matrix form, it is passed as [['Variable']]
        '''
        flag, exprs = try_eval(exprs, l, g)
        #print("after try_eval", flag, exprs)
        if not flag:
            if isinstance(exprs, str):
                exprs = [exprs]
            # elif isinstance(exprs, float):
            #    exprs = [exprs]
            # elif isinstance(exprs, long):
            #    exprs = [exprs]
            elif isinstance(exprs, numbers.Number):
                exprs = [exprs]
            else:
                pass
        if isinstance(exprs, list) and isinstance(exprs[0], list):
            exprs = exprs[0]
        #dprint1("final exprs", exprs)
        self.l = {}
        # TODO g must be copied since some may passs _global_ns.
        # (I don't do it now since it requires a substantial testing)
        self.g = g
        for key in l.keys():
            self.g[key] = l[key]
        self.real = real
        self.variables = []

        self.co = []

        for expr in exprs:
            if isinstance(expr, str):
                code = compile(expr.strip(), '<string>', 'eval')
                names = code.co_names
                for n in names:
                    if (n in g and isinstance(g[n], NativeCoefficientGenBase)):
                        coeff_var = CoefficientVariable(g[n], l, g)
                        self.variables.append((n, coeff_var))
                    elif n in g and isinstance(g[n], Variable):
                        self.variables.append((n, g[n]))
                        for nn in g[n].dependency:
                            self.variables.append((nn, g[nn]))
                    else:
                        pass
                self.co.append(code)
            else:
                self.co.append(expr)

        # 'x, y, z' -> 'x', 'y', 'z'
        self.ind_vars = [x.strip() for x in ind_vars.split(',')]
        self.exprs = exprs
        self.flags = [isinstance(co, types.CodeType) for co in self.co]
        self.variables_dd = dict(self.variables)

    def EvalValue(self, x):
        for k, name in enumerate(self.ind_vars):
            self.l[name] = x[k]
        for n, v in self.variables:
            kwargs = {}
            for nn in v.dependency:
                kwargs[nn] = self.variables_dd[nn]()
            for nn in v.grad:
                kwargs['grad'+nn] = self.variables_dd[nn].eval_grad()
            for nn in v.curl:
                kwargs['curl'+nn] = self.variables_dd[nn].eval_curl()
            for nn in v.div:
                kwargs['div'+nn] = self.variables_dd[nn].eval_div()
            self.l[n] = v(**kwargs)

        val = [eval_code(co, self.g, self.l, flag=flag)
               for co, flag in zip(self.co, self.flags)]
        return np.array(val, copy=False).flatten()
        # return np.array(val, copy=False).ravel()  ## this may be okay


class PhysCoefficient(mfem.PyCoefficient, Coefficient_Evaluator):
    def __init__(self, exprs, ind_vars, l, g, real=True, isArray=False):
        # if not isArray:
        #    exprs = [exprs]
        Coefficient_Evaluator.__init__(self, exprs, ind_vars, l, g, real=real)
        mfem.PyCoefficient.__init__(self)

        handle_allow_python_function_coefficient(
            "Python function coefficient is created")

    def __repr__(self):
        return self.__class__.__name__ + "(PhysCoefficeint)"

    def Eval(self, T, ip):
        for n, v in self.variables:
            v.set_point(T, ip, self.g, self.l)
        return super(PhysCoefficient, self).Eval(T, ip)

    def EvalValue(self, x):
        # set x, y, z to local variable so that we can use in
        # the expression.

        # note that this class could return array, since
        # a user may want to define multiple variables
        # as an array. In such case, subclass should pick
        # one element.
        val = Coefficient_Evaluator.EvalValue(self, x)
        if len(self.co) == 1 and len(val) == 1:
            return val[0]
        return val


class VectorPhysCoefficient(mfem.VectorPyCoefficient, Coefficient_Evaluator):
    def __init__(self, sdim, exprs, ind_vars, l, g, real=True):
        Coefficient_Evaluator.__init__(self, exprs, ind_vars, l, g, real=real)
        mfem.VectorPyCoefficient.__init__(self, sdim)
        self.sdim = sdim

        handle_allow_python_function_coefficient(
            "Python function coefficient is created")

    def __repr__(self):
        return self.__class__.__name__ + "(VectorPhysCoefficeint)"

    def Eval(self, V, T, ip):
        # if V is Matrix, we need to make our own loop to
        # properly set spacial points in variables.
        if isinstance(ip, mfem.IntegrationPoint):
            for n, v in self.variables:
                v.set_point(T, ip, self.g, self.l)
            return super(VectorPhysCoefficient, self).Eval(V, T, ip)
        elif isinstance(ip, mfem.IntegrationRule):
            M = V
            ir = ip
            Mi = mfem.Vector()
            M.SetSize(self.sdim, ir.GetNPoints())
            for k in range(ir.GetNPoints()):
                M.GetColumnReference(k, Mi)
                ip = ir.IntPoint(k)
                for n, v in self.variables:
                    v.set_point(T, ip, self.g, self.l)
                super(VectorPhysCoefficient, self).Eval(Mi, T, ip)

    def EvalValue(self, x):
        return Coefficient_Evaluator.EvalValue(self, x)


class MatrixPhysCoefficient(mfem.MatrixPyCoefficient, Coefficient_Evaluator):
    def __init__(self, sdim, exprs, ind_vars, l, g, real=True):
        self.sdim = sdim
        Coefficient_Evaluator.__init__(self, exprs, ind_vars, l, g, real=real)
        mfem.MatrixPyCoefficient.__init__(self, sdim)

        handle_allow_python_function_coefficient(
            "Python function coefficient is created")

    def __repr__(self):
        return self.__class__.__name__ + "(MatrixPhysCoefficeint)"

    def Eval(self, K, T, ip):
        for n, v in self.variables:
            v.set_point(T, ip, self.g, self.l)
        return super(MatrixPhysCoefficient, self).Eval(K, T, ip)

    def EvalValue(self, x):
        val = Coefficient_Evaluator.EvalValue(self, x)
        # reshape tosquare matrix (not necessariliy = sdim x sdim)
        # if elment is just one, it formats to diagonal matrix
        s = val.size
        if s == 1:
            return np.eye(self.sdim) * val[0]
        else:
            dim = int(np.sqrt(s))
            return val.reshape(dim, dim)


class Phys(Model, Vtable_mixin, NS_mixin):
    hide_ns_menu = True
    hide_nl_panel = False
    dep_vars_base = []
    der_vars_base = []
    der_vars_vec = []

    has_essential = False
    is_secondary_condition = False   # if true, there should be other
    # condtion assigned to the same
    # edge/face/domain

    _has_4th_panel = True
    can_rename = True
    allow_custom_intorder = False

    # matrix element can be time-dependent
    can_timedpendent = False

    vt = Vtable(tuple())
    vt2 = Vtable(tuple())
    vt3 = Vtable(tuple())
    nlterms = []

    def __init__(self, *args, **kwargs):
        super(Phys, self).__init__(*args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def get_info_str(self):
        return NS_mixin.get_info_str(self)

    @property
    def update_flag(self):
        if not hasattr(self, '_update_flag'):
            self._update_flag = False
        return self._update_flag

    @update_flag.setter
    def update_flag(self, prop):
        self._update_flag = prop

    def attribute_set(self, v):
        v = super(Phys, self).attribute_set(v)
        self.vt.attribute_set(v)
        self.vt3.attribute_set(v)

        nl_config = dict()
        for k in self.nlterms:
            nl_config[k] = []
        v['nl_config'] = (False, nl_config)
        v['timestep_config'] = [True, False, False]
        v['timestep_weight'] = ["1", "0", "0"]
        v['isTimeDependent'] = False
        v['isTimeDependent_RHS'] = False
        v['isJacobian'] = False
        v['add_intorder'] = 0
        return v

    def get_possible_bdry(self):
        return []

    def get_possible_domain(self):
        return []

    def get_possible_edge(self):
        return []

    def get_possible_pair(self):
        return []

    def get_possible_point(self):
        return []

    def get_probe(self):
        '''
        return probe name
        '''
        return ''

    def get_independent_variables(self):
        p = self.get_root_phys()
        ind_vars = p.ind_vars
        return [x.strip() for x in ind_vars.split(',')]

    def is_complex(self):
        return False

    def get_restriction_array(self, engine, idx=None):
        mesh = engine.emeshes[self.get_root_phys().emesh_idx]
        intArray = mfem.intArray

        if isinstance(self, Domain):
            size = max(np.max(mesh.attributes.ToList()), engine.max_attr)
        else:
            size = max(
                np.max(
                    mesh.bdr_attributes.ToList()),
                engine.max_bdrattr)

        arr = [0] * size
        if idx is None:
            idx = self._sel_index
        for k in idx:
            arr[k - 1] = 1

        return intArray(arr)

    def restrict_coeff(self, coeff, engine, vec=False, matrix=False,
                       idx=None):
        if len(self._sel_index) == 1 and self._sel_index[0] == -1:
            return coeff
        arr = self.get_restriction_array(engine, idx)
        ret = []
        flag = False

        if not isinstance(coeff, tuple):
            flag = True
            coeff = [coeff]

        for c in coeff:
            if (isinstance(c, mfem.Coefficient) or
                isinstance(c, mfem.VectorCoefficient) or
                    isinstance(c, mfem.MatrixCoefficient)):
                if vec:
                    ret.append(mfem.VectorRestrictedCoefficient(c, arr))
                elif matrix:
                    ret.append(mfem.MatrixRestrictedCoefficient(c, arr))
                else:
                    ret.append(mfem.RestrictedCoefficient(c, arr))
            else:
                ret.append(c)

        if not flag:
            return tuple(ret)
        else:
            return ret[0]

    def get_essential_idx(self, idx):
        '''
        return index of essentail bdr for idx's fespace
        '''
        raise NotImplementedError(
            "you must specify this method in subclass")

    def get_root_phys(self):
        p = self
        while isinstance(p.parent, Phys):
            p = p.parent
        return p

    def get_projection(self):
        return 1

    def update_param(self):
        '''
        called everytime it assembles either matrix or rhs
        '''
        if hasattr(self, 'vt'):
            self.vt.preprocess_params(self)

    def postprocess_extra(self, sol, flag, sol_extra):
        '''
        postprocess extra (Lagrange multiplier) segment
        of solutions.

        sol_extra is a dictionary in which this method
        will add processed data (dataname , value).
        '''
        raise NotImplementedError(
            "you must specify this method in subclass")

    def get_extra_NDoF(self, kfes=0):
        return 0

    def has_extra_DoF(self, kfes=0):
        return False

    def is_extra_RHSonly(self):
        return False

    def has_extra_DoF2(self, kfes, phys, jmatrix):
        '''
        subclass has to overwrite this if extra DoF can couple
        with other FES.
        '''
        if not self.check_jmatrix(jmatrix):
            return False

        if (phys == self.get_root_phys()):
            # if module set only has_extra_DoF, the extra variable
            # couples only in the same Phys module and jmatrix == 0

            # ToDo. Should add deprecated message
            return self.has_extra_DoF(kfes=kfes)
        else:
            return False

    def extra_DoF_name(self):
        '''
        default DoF name
        '''
        return self.get_root_phys().dep_vars[0] + '_' + self.name()

    def extra_DoF_name2(self, kfes=0):
        '''
        default DoF name
        '''
        return self.extra_DoF_name()

    def check_extra_update(self, mode):
        '''
        mode = 'B' or 'M'
        'M' return True, if M needs to be updated
        'B' return True, if B needs to be updated
        default, return true for B not M
        '''
        if self._update_flag:
            if mode == 'B':
                return True
        return False

    def has_bf_contribution(self, kfes):
        return False

    def has_bf_contribution2(self, kfes, jmatrix):
        '''
        subclass has to overwrite this if extra DoF can couple
        with other FES.
        '''
        if not self.check_jmatrix(jmatrix):
            return False
        return self.has_bf_contribution(kfes)

    def has_lf_contribution(self, kfes):
        return False

    def has_lf_contribution2(self, kfes, jmatrix):
        '''
        subclass has to overwrite this if extra DoF can couple
        with other FES.
        '''
        if not self.check_jmatrix(jmatrix):
            return False
        return self.has_lf_contribution(kfes)

    def has_interpolation_contribution(self, kfes):
        return False

    def has_mixed_contribution(self):
        return False

    def has_mixed_contribution2(self, jmatrix):
        '''
        subclass has to overwrite this if extra DoF can couple
        with other FES.
        '''
        if not self.check_jmatrix(jmatrix):
            return False
        return self.has_mixed_contribution()

    def has_aux_op(self, phys1, kfes1, phys2, kfes2):
        return False

    def has_aux_op2(self, phys1, kfes1, phys2, kfes2, jmatrix):
        if not self.check_jmatrix(jmatrix):
            return False
        return self.has_aux_op(phys1, kfes1, phys2, kfes2)

    '''
    def set_matrix_weight(self, get_matrix_weight):
        # self._mat_weight = get_matrix_weight(self.timestep_config,
        #                                      self.timestep_weight)
        w = get_matrix_weight(self.timestep_config)
        if sum(w) > 0:
           self._mat_weight = w
    '''

    def set_matrix_weight(self, w):
        '''
        matrix weight = [y, dy/dt, dy/dt^2, Jac(y), Jac(dy/dt), Jac(dy/dt^2)
        '''
        from petram.engine import max_matrix_num
        ww = [0]*max_matrix_num

        for i, val in enumerate(w):
            ww[i] = val
        if self.isJacobian:
            ww[:(max_matrix_num//2)] = [False]*(max_matrix_num//2)
        else:
            ww[(max_matrix_num//2):] = [False]*(max_matrix_num//2)
        self._mat_weight = ww

    def get_matrix_weight(self):
        return self._mat_weight

    def check_jmatrix(self, jmatrix):
        if not hasattr(self, "_mat_weight"):
            # _mat_weight is "usually" set by solver before
            # if it is not set, we are not solving this physics
            return False

        return self._mat_weight[jmatrix] != 0

    def get_mixedbf_loc(self):
        '''
        r, c, and flag of MixedBilinearForm
        flag -1 :  use conjugate when building block matrix

        if r, c is int, it is offset in the physics module, a couple between
        variables in the same module can use this mode. Otherwise, r and c
        are the name of variables
        '''
        return []

    def add_bf_contribution(self, engine, a, real=True, **kwargs):
        raise NotImplementedError(
            "you must specify this method in subclass")

    def add_lf_contribution(self, engine, b, real=True, **kwargs):
        raise NotImplementedError(
            "you must specify this method in subclass")

    def add_extra_contribution(self, engine, kfes, **kwargs):
        '''
        return four elements
        M12, M21, M22, rhs
        '''
        raise NotImplementedError(
            "you must specify this method in subclass")

    def add_interplation_contribution(self, engine, **kwargs):
        '''
        P^t A P y = P^t f, x = Py
        return P, nonzero, zero diagonals.
        '''
        raise NotImplementedError(
            "you must specify this method in subclass")

    def apply_essential(self, engine, gf, **kwargs):
        raise NotImplementedError(
            "you must specify this method in subclass")

    def get_init_coeff(self, engine, **kwargs):
        return None

    def apply_init(self, engine, gf, **kwargs):
        import warnings
        warnings.warn("apply init is not implemented to " + str(self))

    def add_mix_contribution(self, engine, a, r, c, is_trans, real=True):
        '''
        return array of crossterms
        [[vertical block elements], [horizontal block elements]]
        array length must be the number of fespaces
        for the physics

        r, c : r,c of block matrix
        is_trans: indicate if transposed matrix is filled
        '''
        raise NotImplementedError(
            "you must specify this method in subclass")

    def add_mix_contribution2(self, engine, a, r, c,
                              is_trans, is_conj, real=True):
        return self.add_mix_contribution(engine, a, r, c, is_trans, real=real)

    def add_variables(self, solvar, n, solr, soli=None):
        '''
        add model variable so that a user can interept simulation
        results. It is also used in cross-physics interaction.

        '''
        pass

    def add_domain_variables(self, v, n, suffix, ind_vars):
        pass

    def add_bdr_variables(self, v, n, suffix, ind_vars):
        pass

    def panel1_param(self):
        return self.vt.panel_param(self)

    def get_panel1_value(self):
        return self.vt.get_panel_value(self)

    def preprocess_params(self, engine):
        self.vt.preprocess_params(self)
        self.vt3.preprocess_params(self)
        return

    def import_panel1_value(self, v):
        return self.vt.import_panel_value(self, v)

    def panel1_tip(self):
        return self.vt.panel_tip()

    def panel3_param(self):
        from petram.pi.widget_nl import NonlinearTermPanel
        l = self.vt3.panel_param(self)
        if self.hide_nl_panel or len(self.nlterms) == 0:
            return l
        setting = {'UI': NonlinearTermPanel, 'names': self.nlterms}
        l.append([None, None, 99, setting])
        return l

    def get_panel3_value(self):
        if self.hide_nl_panel or len(self.nlterms) == 0:
            return self.vt3.get_panel_value(self)
        else:
            return self.vt3.get_panel_value(self) + [self.nl_config]

    def import_panel3_value(self, v):
        if self.hide_nl_panel or len(self.nlterms) == 0:
            self.vt3.import_panel_value(self, v)
        else:
            self.vt3.import_panel_value(self, v[:-1])
            self.nl_config = v[-1]

    def panel3_tip(self):
        if self.hide_nl_panel or len(self.nlterms) == 0:
            return self.vt3.panel_tip()
        else:
            return self.vt3.panel_tip() + [None]

    def panel4_param(self):
        setting = {"text": ' '}
        if self.has_essential:
            ll = [['', False, 3, {"text": "Time dependent"}], ]

        else:
            ll = [['y(t)', True, 3, {"text": ""}],
                  ['dy/dt', False, 3, {"text": ""}],
                  ['d2y/dt2', False, 3, {"text": ""}],
                  ['Gradient', False, 3, {"text": ""}],
                  ['Varying (in time/for loop) Term.', False, 3, {"text": ""}], ]
        if self.allow_custom_intorder:
            ll.append(['Increase int. order', '0', 400, ''])
        return ll

    def panel4_tip(self):
        return None

    def import_panel4_value(self, value):
        if self.has_essential:
            self.isTimeDependent = value[0]

        else:
            self.timestep_config[0] = value[0]
            self.timestep_config[1] = value[1]
            self.timestep_config[2] = value[2]
            self.isJacobian = value[3]
            self.isTimeDependent = value[4]

        if self.allow_custom_intorder:
            self.add_intorder = int(value[-1])

    def get_panel4_value(self):
        if self.has_essential:
            ret = [self.isTimeDependent]
        else:
            ret = self.timestep_config[0:3] + \
                [self.isJacobian, self.isTimeDependent]
        if self.allow_custom_intorder:
            ret = ret + [self.add_intorder]
        return ret

    @property
    def geom_dim(self):
        root_phys = self.get_root_phys()
        return root_phys.geom_dim

    @property
    def dim(self):
        root_phys = self.get_root_phys()
        return root_phys.dim

    @property
    def integrator_realimag_mode(self):
        if hasattr(self, "_realimag_mode"):
            return self._realimag_mode
        else:
            return ""

    def set_integrator_realimag_mode(self, value):
        if value:
            self._realimag_mode = "real"
        else:
            self._realimag_mode = "imag"

    def process_complex_coefficient(self, coeff):
        if not hasattr(coeff[0], "get_real_coefficient"):
            return coeff

        if self.integrator_realimag_mode == "real":
            return sum(((coeff[0].get_real_coefficient(),), coeff[1:]), ())
        elif self.integrator_realimag_mode == "imag":
            return sum(((coeff[0].get_imag_coefficient(),), coeff[1:]), ())
        else:
            return coeff

    def add_integrator(self, engine, name, coeff, adder, integrator, idx=None, vt=None,
                       transpose=False, ir=None):
        if coeff is None:
            return
        if vt is None:
            vt = self.vt
        # if vt[name].ndim == 0:
        if not isinstance(coeff, tuple):
            coeff = (coeff, )

        coeff = self.process_complex_coefficient(coeff)

        if coeff[0] is None:
            return
        elif isinstance(coeff[0], mfem.Coefficient):
            coeff = self.restrict_coeff(coeff, engine, idx=idx)
        elif isinstance(coeff[0], mfem.VectorCoefficient):
            coeff = self.restrict_coeff(coeff, engine, vec=True, idx=idx)
        elif isinstance(coeff[0], mfem.MatrixCoefficient):
            coeff = self.restrict_coeff(coeff, engine, matrix=True, idx=idx)
        else:
            assert False, "Unknown coefficient type: " + str(type(coeff[0]))

        itg = integrator(*coeff)
        itg._linked_coeff = coeff  # make sure that coeff is not GCed.

        if transpose:
            itg2 = mfem.TransposeIntegrator(itg)
            itg2._link = itg
            if ir is not None:
                itg2.SetIntRule(ir)
            adder(itg2)
        else:
            if ir is not None:
                itg.SetIntRule(ir)
            adder(itg)

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys', self)

    #
    #  utilities for add_domain_variables
    #

    def do_add_scalar_expr(self, v, suffix, ind_vars, name, f_name,
                           add_diag=False):

        from petram.helper.variables import add_expression, add_constant
        from petram.helper.variables import NativeCoefficientGenBase

        kywds = {}
        if isinstance(self, Domain):
            kywds['domains'] = self._sel_index
            kywds['gdomain'] = self._global_ns
        elif isinstance(self, Bdry):
            kywds['bdrs'] = self._sel_index
            kywds['gbdr'] = self._global_ns

        if isinstance(f_name, NativeCoefficientGenBase):
            pass
        elif isinstance(f_name, str):
            add_expression(v, name, suffix, ind_vars, f_name,
                           [], **kywds)

        else:
            add_constant(v, name, suffix, f_name, **kywds)

        if add_diag:
            '''
            convension:
               scalar parameters should be s+name

               ex) in RF epsilonr is (3,3)
                  sepsilonr : scarlar version
                  epsilonr : matrix
            '''
            size = int(add_diag)
            add_expression(v, name[1:], suffix, ind_vars,
                           'diag([1]*' + str(size) + ')*' + name,
                           [name], **kywds)

    def do_add_matrix_expr(self, v, suffix, ind_vars, name, f_name):

        from petram.helper.variables import add_expression, add_constant
        from petram.helper.variables import NativeCoefficientGenBase

        kywds = {}
        if isinstance(self, Domain):
            kywds['domains'] = self._sel_index
            kywds['gdomain'] = self._global_ns
        elif isinstance(self, Bdry):
            kywds['bdrs'] = self._sel_index
            kywds['gbdr'] = self._global_ns

        if isinstance(f_name, NativeCoefficientGenBase):
            pass
        elif len(f_name) == 1:
            if not isinstance(f_name[0], str):
                expr = f_name[0].__repr__()
            else:
                expr = f_name[0]
            add_expression(v, name, suffix, ind_vars, expr,
                           [], **kywds)
        else:  # elemental format
            expr_txt = [x.__repr__() if not isinstance(x, str) else x
                        for x in f_name]
            a = '[' + ','.join(expr_txt[:3]) + ']'
            b = '[' + ','.join(expr_txt[3:6]) + ']'
            c = '[' + ','.join(expr_txt[6:]) + ']'
            expr = '[' + ','.join((a, b, c)) + ']'
            add_expression(v, name, suffix, ind_vars, expr,
                           [], **kywds)

    def do_add_matrixlike_component_expr_scalar(
            self, v, suffix, ind_vars, var, name):
        '''
        this routine add matrix-like access to scalar variable

           epsilonr1xx = epsilonr1
           epsilonr1yy = epsilonr1
           epsilonr1zz = epsilonr1
           the reset is defined as 0
        '''

        from petram.helper.variables import add_component_expression as addc_expression
        from itertools import product

        kywds = {}
        if isinstance(self, Domain):
            kywds['domains'] = self._sel_index
            kywds['gdomain'] = self._global_ns
        elif isinstance(self, Bdry):
            kywds['bdrs'] = self._sel_index
            kywds['gbdr'] = self._global_ns

        ll = len(var)
        for lll in product(var, var):
            componentname = lll[0] + lll[1]
            if lll[0] == lll[1]:
                addc_expression(v, name, suffix, ind_vars, name, [name], componentname,
                                **kywds)
            else:
                addc_expression(v, name, suffix, ind_vars, '0', [], componentname,
                                **kywds)

    def do_add_matrix_component_expr(self, v, suffix, ind_vars, var, name):
        '''
        this routine add matrix-like access to scalar variable

           epsilonr1xx = epsilonr1[0,0]
           epsilonr1yy = epsilonr1[1,1]

        var
        '''

        from petram.helper.variables import add_component_expression as addc_expression
        from itertools import product

        kywds = {}
        if isinstance(self, Domain):
            kywds['domains'] = self._sel_index
            kywds['gdomain'] = self._global_ns
        elif isinstance(self, Bdry):
            kywds['bdrs'] = self._sel_index
            kywds['gbdr'] = self._global_ns

        ll = len(var)
        for lll in product(var, var):
            componentname = lll[0] + lll[1]
            i1 = var.index(lll[0])
            i2 = var.index(lll[1])
            expr = name + '[' + str(i1) + ',' + str(i2) + ']'
            addc_expression(v, name, suffix, ind_vars, expr, [name], componentname,
                            **kywds)

    def get_coefficient_from_expression(
            self, c, cotype, use_dual=False, real=True, is_conj=False):
        from petram.phys.coefficient import SCoeff, VCoeff, DCoeff, MCoeff

        if self.get_root_phys().vdim > 1:
            dim = self.get_root_phys().vdim
        else:
            dim = self.get_root_phys().geom_dim

        if cotype == 'S':
            # for b in self.itg_choice():
            #   if b[0] == self.integrator: break
            # if not "S*2" in b[3]:
            if not use_dual:
                c_coeff = SCoeff(c, self.get_root_phys().ind_vars,
                                 self._local_ns, self._global_ns,
                                 real=real, conj=is_conj)
            else:  # so far this is only for an elastic integrator
                c_coeff = (SCoeff(c, self.get_root_phys().ind_vars,
                                  self._local_ns, self._global_ns,
                                  real=real, conj=is_conj, component=0),
                           SCoeff(c, self.get_root_phys().ind_vars,
                                  self._local_ns, self._global_ns,
                                  real=real, conj=is_conj, component=1))
        elif cotype == 'V':
            c_coeff = VCoeff(dim, c, self.get_root_phys().ind_vars,
                             self._local_ns, self._global_ns,
                             real=real, conj=is_conj)
        elif cotype == 'M':
            c_coeff = MCoeff(dim, c, self.get_root_phys().ind_vars,
                             self._local_ns, self._global_ns,
                             real=real, conj=is_conj)
        elif cotype == 'D':
            c_coeff = DCoeff(dim, c, self.get_root_phys().ind_vars,
                             self._local_ns, self._global_ns,
                             real=real, conj=is_conj)
        return c_coeff

    def compile_coeffs(self, *kargs):
        '''
        jit compile coefficient
        '''
        pass


data = [("order", VtableElement("order", type='int',
                                guilabel="order", no_func=True,
                                default=1, tip="element order",))]


class PhysModule(Phys):
    hide_ns_menu = False
    dim_fixed = True  # if ndim of physics is fixed
    can_rename = False
    vt_order = Vtable(data)

    def __setstate__(self, state):
        Phys.__setstate__(self, state)
        if self.sel_index == 'all':
            self.sel_index = ['all']

    def get_info_str(self):
        txt = self.dep_vars
        if NS_mixin.get_info_str(self) != "":
            txt.append(NS_mixin.get_info_str(self))
        return ",".join(txt)

    def preprocess_params(self, engine):
        self.vt_order.preprocess_params(self)

    @property
    def geom_dim(self):  # dim of geometry
        return len(self.ind_vars.split(','))

    @property
    def dim(self):  # dim of FESpace independent variable
        return self.ndim

    @dim.setter
    def dim(self, value):
        self.ndim = value

    @property
    # mesh_idx is dynamic value to point a mesh index used in solve.
    def emesh_idx(self):
        return self._emesh_idx

    @emesh_idx.setter
    def emesh_idx(self, value):
        self._emesh_idx = value

    def goem_signature(self):
        pass

    @abstractmethod
    def get_fec_type(self, idx):
        pass
        '''
        return FEC type
        H1
        H1v2 (vector dim)
        ND
        RT
        '''

    def fes_order(self, idx):
        self.vt_order.preprocess_params(self)
        return self.order

    def attribute_set(self, v):
        v = super(PhysModule, self).attribute_set(v)
        v["vdim"] = 1
        v["element"] = 'H1_FECollection'
        v["ndim"] = 2  # logical dimension (= ndim of mfem::Mesh)
        v["ind_vars"] = 'x, y'
        v["dep_vars_suffix"] = ''
        v["mesh_idx"] = 0
        v['sel_index'] = ['all']
        v['sel_index_txt'] = 'all'

        # subclass can use this flag to generate FESpace for time derivative
        # see WF_model
        v['generate_dt_fespace'] = False
        v = self.vt_order.attribute_set(v)
        return v

    def get_dependent_variables(self):
        from petram.helper.variables import append_suffix_to_expression
        cls = self.__class__
        suffix = self.dep_vars_suffix
        vecs = cls.der_vars_vec
        names = [append_suffix_to_expression(x, vecs, suffix)
                 for x in cls.der_vars_base]
        return names

    def onVarNameChanged(self, evt):
        evt.GetEventObject().TopLevelParent.OnRefreshTree()

    def get_var_suffix_var_name_panel(self):
        from petram.pi.widgets import TextCtrlCallBack
        a = ["dep. vars. suffix", self.dep_vars_suffix, 99, {'UI': TextCtrlCallBack,
                                                             'callback_on_enter': True,
                                                             'callback_method': self.onVarNameChanged}]
        b = ["dep. vars.", ','.join(self.dep_vars_base), 99, {'UI': TextCtrlCallBack,
                                                              'callback_on_enter': True,
                                                              'callback_method': self.onVarNameChanged}]

        return a, b

    def panel1_param(self):
        ll = self.vt_order.panel_param(self)
        return [["mesh num.", self.mesh_idx, 400, {}],
                ["element", self.element, 2, {}]] + ll

    def panel1_tip(self):
        ll = self.vt_order.panel_tip()
        return ["index of mesh", "element type"] + ll

    def get_panel1_value(self):
        ll = self.vt_order.get_panel_value(self)
        return [self.mesh_idx, self.element] + ll

    def import_panel1_value(self, v):
        self.mesh_idx = int(v[0])
        self.element = str(v[1])
        self.vt_order.import_panel_value(self, [v[2]])
        return v[3:]

    def panel2_param(self):
        import wx

        if self.geom_dim == 3:
            choice = ("Volume", "Surface", "Edge")
        elif self.geom_dim == 2:
            choice = ("Surface", "Edge")
        elif self.geom_dim == 1:
            choice = ("Edge", )

        from petram.model import validate_sel

        if self.dim_fixed:
            return [["index", 'all', 0, {'changing_event': True,
                                         'setfocus_event': True,
                                         'validator': validate_sel,
                                         'validator_param': self}, ]]

        else:
            p = ["Type", choice[0], 4,
                 {"style": wx.CB_READONLY, "choices": choice}]
            return [p, ["index", 'all', 0, {'changing_event': True,
                                            'setfocus_event': True,
                                            'validator': validate_sel,
                                            'validator_param': self}, ]]

    def get_panel2_value(self):
        choice = ["Point", "Edge", "Surface", "Volume", ]
        if self.dim_fixed:
            return (self.sel_index_txt, )
        else:
            return choice[self.dim], self.sel_index_txt

    def import_panel2_value(self, v):
        if self.dim_fixed:
            self.sel_index_txt = str(v[0])
        else:
            if str(v[0]) == "Volume":
                self.dim = 3
            elif str(v[0]) == "Surface":
                self.dim = 2
            elif str(v[0]) == "Edge":
                self.dim = 1
            else:
                self.dim = 1
            self.sel_index_txt = str(v[1])

        self.update_dom_selection()
        return True

    @property
    def dep_vars(self):
        raise NotImplementedError(
            "you must specify this method in subclass")

    @property
    def dep_vars_base(self, name):
        raise NotImplementedError(
            "you must specify this method in subclass")

    @property
    def dep_vars0(self):
        # solver uses this method to find dep_vars to construct linear system
        # used (for example) to omit d/dt term from FESspace variables
        raise NotImplementedError(
            "you must specify this method in subclass")

    def dep_var_index(self, name):
        return self.dep_vars.index(name)

    def check_new_fespace(fespaces, meshes):
        mesh = meshes[self.mesh_name]
        fespacs = fespeces[self]

    def get_fec(self):
        name = self.dep_vars
        return [(name[0], self.element)]

    def is_complex(self):
        return False

    _possible_constraints = None

    @classmethod
    def _set_possible_constraints(cls, name):
        '''
        utility to define a possible constraints dynamically expandable.
        '''
        from petram.helper.phys_module_util import get_phys_constraints
        constraints = get_phys_constraints(name)
        cls._possible_constraints = constraints

    def get_possible_domain(self):
        from petram.phys.wf.wf_constraints import WF_WeakDomainBilinConstraint, WF_WeakDomainLinConstraint
        return [WF_WeakDomainBilinConstraint, WF_WeakDomainLinConstraint]

    def get_possible_bdry(self):
        from petram.phys.wf.wf_constraints import WF_WeakBdryBilinConstraint, WF_WeakBdryLinConstraint
        return [WF_WeakBdryBilinConstraint, WF_WeakBdryLinConstraint]

    def get_possible_point(self):
        from petram.phys.wf.wf_constraints import (WF_WeakPointBilinConstraint,
                                                   WF_WeakPointLinConstraint)

        # (note) BilinearForm does not suppoert delta coefficent
        return [WF_WeakPointBilinConstraint, WF_WeakPointLinConstraint]

    def get_possible_child(self):
        from petram.phys.aux_variable import AUX_Variable
        from petram.phys.aux_operator import AUX_Operator
        return [AUX_Variable, AUX_Operator]

    def get_possible_pair(self):
        from petram.phys.projection import BdrProjection, DomainProjection
        return [DomainProjection, BdrProjection, ]

    def soldict_to_solvars(self, soldict, variables):
        keys = list(soldict)
        depvars = self.dep_vars
        suffix = self.dep_vars_suffix
        ind_vars = [x.strip() for x in self.ind_vars.split(',')]

        for k in keys:
            n = k.split('_')[0]
            if n in depvars:
                sol = soldict[k]
                solr = sol[0]
                soli = sol[1] if len(sol) > 1 else None
                self.add_variables(variables, n, solr, soli)

        # collect all definition (domain specific expressions) from children
        for mm in self.walk():
            if not mm.enabled:
                continue
            if mm is self:
                continue

            mm.add_domain_variables(variables, n, suffix, ind_vars)
            mm.add_bdr_variables(variables, n, suffix, ind_vars)


    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys', self)

    def is_viewmode_grouphead(self):
        return True

    def get_mesh_ext_info(self, m):
        from petram.mesh.mesh_extension import MeshExtInfo

        info = MeshExtInfo(dim=self.dim, base=self.mesh_idx)

        if len(self.sel_index) == 0:
            return None
        if self.sel_index[0] != 'all':
            info.set_selection(self.sel_index)
        else:
            if self.dim == 3:
                info.set_selection(list(m.extended_connectivity['vol2surf']))
            elif self.dim == 2:
                info.set_selection(list(m.extended_connectivity['surf2line']))
            else:
                info.set_selection(list(m.extended_connectivity['vert2vert']))

        return info

    def update_dom_selection(self, all_sel=None):
        from petram.model import convert_sel_txt
        try:
            arr = convert_sel_txt(self.sel_index_txt, self._global_ns)
            self.sel_index = arr
        except BaseException:
            assert False, "failed to convert " + self.sel_index_txt

        if all_sel is None:
            # clinet GUI panel operation ends here
            return

        allv, alls, alle = all_sel
        if len(self.sel_index) != 0 and self.sel_index[0] == 'all':
            if self.dim == 3:
                self.sel_index = allv
            if self.dim == 2:
                self.sel_index = alls
            if self.dim == 1:
                self.sel_index = alle

    def collect_probes(self):
        probes = [mm.get_probe() for mm in self.walk() if mm.is_enabled()]
        probes = [x for x in probes if len(x) != 0]
        txt = ','.join(probes)
        return txt

    def get_dom_bdr_pnt_choice(self, mesh):

        from collections import defaultdict

        def get_internal_bdr(d, dom_choice):
            dd = []
            for k in d:
                if k in dom_choice:
                    dd.extend(d[k])
            seen = defaultdict(int)
            for k in dd:
                seen[k] += 1
            return [k for k in seen if seen[k] > 1]

        if self.dim == 3:
            kk = 'vol2surf'
        elif self.dim == 2:
            kk = 'surf2line'
        elif self.dim == 1:
            kk = 'line2vert'
        else:
            assert False, "not supported"
        d = mesh.extended_connectivity[kk]

        if d is None:
            return [], [], [], []

        dom_choice = list(d)
        bdr_choice = list(set(np.hstack([list(d[x]) for x in d])))
        pnt_choice = list(mesh.extended_connectivity["vert2vert"])

        if self.sel_index[0] != 'all':
            dom_choice = [int(x) for x in self.sel_index]
            bdrs = [d[int(x)] if int(x) in d else [] for x in self.sel_index]
            bdr_choice = list(np.unique(np.hstack(bdrs)).astype(int))
        internal_bdr = get_internal_bdr(d, dom_choice)

        from petram.mfem_config import use_parallel
        if use_parallel:
            from mfem.common.mpi_debug import nicePrint
            from petram.helper.mpi_recipes import allgather
            dom_choice = list(set(sum(allgather(dom_choice), [])))
            bdr_choice = list(set(sum(allgather(bdr_choice), [])))
            internal_bdr = list(set(sum(allgather(internal_bdr), [])))

        # return unique list
        return list(set(dom_choice)), list(
            set(bdr_choice)), list(set(pnt_choice)), list(set(internal_bdr)),
        # return dom_choice, bdr_choice



from petram.model import Model
from petram.helper.variables import var_g
import traceback
from petram.namespace_mixin import NSRef_mixin

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('PP_Model')

ll = var_g.copy()


class PostProcessBase(Model):
    @property
    def _global_ns(self):
        # used for text box validator
        p = self
        while True:
            if isinstance(p, NSRef_mixin):
                break
            p = p.parent
            if p is None:
                # it should not come here...x
                return {}
        return p.find_ns_by_name()

    def run_postprocess(self, engin):
        raise NotImplemented("Subclass must implement run_postprocess")

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('phys')

    def soldict_to_solvars(self, soldict, variables):
        pass

    def update_dom_selection(self, all_sel=None):
        from petram.model import convert_sel_txt
        try:
            arr = convert_sel_txt(self.sel_index_txt, self._global_ns)
            self.sel_index = arr
        except:
            assert False, "failed to convert "+self.sel_index_txt

        if all_sel is None:
            # clinet GUI panel operation ends here
            return

        allv, alls, alle = all_sel
        if len(self.sel_index) != 0 and self.sel_index[0] == 'all':
            if self.sdim == 3:
                self.sel_index = allv
            if self.sdim == 2:
                self.sel_index = alls
            if self.sdim == 1:
                self.sel_index = alle


class PostProcess(PostProcessBase, NSRef_mixin):
    has_2nd_panel = False

    def __init__(self, *args, **kwargs):
        super(PostProcess, self).__init__(*args, **kwargs)
        NSRef_mixin.__init__(self, *args, **kwargs)

    def attribute_set(self, v):
        v = super(PostProcess, self).attribute_set(v)
        v['use_scanner'] = 0
        v['scanner'] = 'Scan("a", [1,2,3])'
        return v

    def panel1_param(self):
        ret = [self.make_param_panel('scanner',  self.scanner), ]
        value = [self.scanner, ]
        return [[None, [False, value], 27, [{'text': 'Use parametric scan'},
                                            {'elp': ret}]], ]

    def get_panel1_value(self):
        v1 = (self.use_scanner, [self.scanner, ])
        return [v1, ]

    def import_panel1_value(self, v):
        v1 = v[0]
        self.use_scanner = v1[0]
        self.scanner = v1[1][0]

    def get_info_str(self):
        txt = []
        if NSRef_mixin.get_info_str(self) != "":
            txt.append(NSRef_mixin.get_info_str(self))
        return ",".join(txt)

    def get_possible_child(self):
        from petram.postprocess.project_solution import DerivedValue
        from petram.postprocess.discrt_v_integration import (LinearformIntegrator,
                                                             BilinearformIntegrator)
        from petram.postprocess.discrt_v_interpolator import Grad, Curl, Div

        return [DerivedValue, LinearformIntegrator, BilinearformIntegrator, Grad, Curl, Div]

    def get_possible_child_menu(self):
        from petram.postprocess.project_solution import DerivedValue
        from petram.postprocess.discrt_v_integration import (LinearformIntegrator,
                                                             BilinearformIntegrator)
        from petram.postprocess.discrt_v_interpolator import Grad, Curl, Div

        return [("", DerivedValue),
                ("Integrator", LinearformIntegrator),
                ("!", BilinearformIntegrator),
                ("Derivative", Grad),
                ("", Curl),
                ("!", Div),
                ]

    def run_postprocess(self, engine):
        dprint1("running postprocess:" + self.name())

    def get_scanner(self, nosave=False):
        try:
            scanner = self.eval_param_expr(str(self.scanner),
                                           'scanner')[0]
            scanner.set_data_from_model(self.root())
        except:
            traceback.print_exc()
            return
        scanner.set_phys_models([self.find_nsobj_by_name(), ])
        return scanner

    def run(self, engine):

        scanner = self.get_scanner() if self.use_scanner else None
        for mm in self.walk():
            if not mm.enabled:
                continue
            if mm is self:
                continue
            if scanner is not None:
                engine.ppname_postfix = ''
                for kcase, case in enumerate(scanner):
                    engine.ppname_postfix = '_'+str(kcase)
                    mm.run_postprocess(engine)
                engine.ppname_postfix = ''
            else:
                engine.ppname_postfix = ''
                mm.run_postprocess(engine)

    # parameters with validator
    def check_param_expr(self, value, param, ctrl):
        try:
            self.eval_param_expr(str(value), param)
            return True
        except:
            import petram.debug
            import traceback
            if petram.debug.debug_default_level > 2:
                traceback.print_exc()
            return False

    def eval_param_expr(self, value, param):
        x = eval(value, self._global_ns)
        dprint2('Value Evaluation ', param, '=', x)
        return x, None

    # note that physics modules overwrite this with more capablie version
    def make_param_panel(self, base_name, value):
        return [base_name + "(=)",  value, 0,
                {'validator': self.check_param_expr,
                 'validator_param': base_name}]



'''
   NCEdgeEvaluator: not-continous edge evaluator
'''
from petram.sol.bdr_nodal_evaluator import get_emesh_idx
from petram.sol.evaluator_agent import EvaluatorAgent
import numpy as np
import weakref
import six

from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
    from mfem.par import GlobGeometryRefiner as GR
else:
    import mfem.ser as mfem
    from mfem.ser import GlobGeometryRefiner as GR

Geom = mfem.Geometry()


def eval_on_edges(obj, expr, solvars, phys):
    '''
    evaluate nodal valus based on preproceessed 
    geometry data

    to be done : obj should be replaced by a dictionary
    '''
    from petram.helper.variables import (Variable,
                                         var_g,
                                         NativeCoefficientGenBase,
                                         CoefficientVariable,
                                         NumbaCoefficientVariable,)

    if len(obj.ifaces) == 0:
        return None
    variables = []

    code = compile(expr, '<string>', 'eval')
    names = code.co_names

    g = {}

    for key in phys._global_ns.keys():
        g[key] = phys._global_ns[key]
    for key in solvars.keys():
        g[key] = solvars[key]

    ll_name = []
    ll_value = []
    var_g2 = var_g.copy()

    new_names = []
    name_translation = {}

    all_names = list(names[:])

    def get_names(names):
        for n in names:
            if (n in g and isinstance(g[n], Variable)):
                new_names = g[n].get_names()
                for x in new_names:
                    all_names.append(x)
                get_names(new_names)
    get_names(names)

    for n in all_names:
        if (n in g and isinstance(g[n], NativeCoefficientGenBase)):
            g[n+"_coeff"] = CoefficientVariable(g[n], g)
            new_names.append(n+"_coeff")
            name_translation[n+"_coeff"] = n

        elif (n in g and isinstance(g[n], NumbaCoefficientVariable)):
            ind_vars = [xx.strip() for xx in phys.ind_vars.split(',')]
            g[n].set_coeff(ind_vars, g)
            new_names.append(n)
            name_translation[n] = n

        elif (n in g and isinstance(g[n], Variable)):
            new_names.extend(g[n].dependency)
            new_names.append(n)
            name_translation[n] = n

        elif n in g:
            new_names.append(n)
            name_translation[n] = n

    for n in new_names:
        if (n in g and isinstance(g[n], Variable)):
            if not g[n] in obj.knowns:
                obj.knowns[g[n]] = (
                    g[n].ncedge_values(ifaces=obj.ifaces,
                                       irs=obj.irs,
                                       gtypes=obj.gtypes,
                                       locs=obj.ptx,
                                       attr1=obj.elattr1,
                                       attr2=obj.elattr2,
                                       g=g, knowns=obj.knowns,
                                       mesh=obj.mesh()[obj.emesh_idx]))

            ll_name.append(name_translation[n])
            ll_value.append(obj.knowns[g[n]])

        elif (n in g):
            var_g2[n] = g[n]

    if len(ll_value) > 0:
        val = np.array([eval(code, var_g2, dict(zip(ll_name, v)))
                        for v in zip(*ll_value)])
    else:
        # if expr does not involve Varialbe, evaluate code once
        # and generate an array
        val = np.array([eval(code, var_g2)]*len(obj.ptx))

    return val


class NCEdgeEvaluator(EvaluatorAgent):
    def __init__(self, battrs, **kwargs):
        super(NCEdgeEvaluator, self).__init__()
        self.battrs = battrs
        self.refine = -1
        self.decimate = kwargs.pop("decimate", 1)

    def preprocess_geometry(self, battrs, emesh_idx=0, decimate=1):
        # we will ignore deciamte for a moment

        mesh = self.mesh()[emesh_idx]
        self.battrs = battrs
        self.knowns = WKD()
        self.iverts = []
        self.ifaces = []

        if mesh.Dimension() == 2:
            def getface(x): return (mesh.GetBdrElementEdges(x)[0][0], 1)
            gettrans = mesh.GetBdrElementTransformation
            getarray = mesh.GetBdrArray
            getelement = mesh.GetBdrElement
            getbasegeom = mesh.GetBdrElementBaseGeometry
            getvertices = mesh.GetBdrElementVertices
            getattr1 = mesh.GetBdrAttribute
            def getattr2(x): return -1

        elif mesh.Dimension() == 1:
            def getface(x): return (x, 1)
            gettrans = mesh.GetElementTransformation
            getarray = mesh.GetDomainArray
            getelement = mesh.GetElement
            getbasegeom = mesh.GetElementBaseGeometry
            getvertices = mesh.GetElementVertices
            getattr1 = mesh.GetAttribute
            def getattr2(x): return -1
        else:
            assert False, "NCEdge Evaluator is not supported for this dimension"

        x = [getarray(battr) for battr in battrs]
        if np.sum([len(xx) for xx in x]) == 0:
            return

        ibdrs = np.hstack(x).astype(int).flatten()
        self.ibeles = np.array(ibdrs)

        ptx = []
        data = []
        ridx = []
        ifaces = []
        self.gtypes = np.zeros(len(self.ibeles), dtype=int)
        self.elattr1 = np.zeros(len(self.ibeles), dtype=int)
        self.elattr2 = np.zeros(len(self.ibeles), dtype=int)

        self.irs = {}

        gtype_st = -1
        nele = 0

        for k, i in enumerate(self.ibeles):
            verts = getvertices(i)
            gtype = getbasegeom(i)
            iface, ort = getface(i)
            #Trs = mesh.GetEdgeElementTransformatio(iface)

            if gtype != gtype_st:
                RefG = GR.Refine(gtype, self.refine)
                ir = RefG.RefPts
                npt = ir.GetNPoints()
                ele = np.array(RefG.RefGeoms.ToList()).reshape(-1, len(verts))
                gtype_st = gtype
                self.irs[gtype] = ir

            T = gettrans(i)
            pt = np.vstack([T.Transform(ir.IntPoint(j)) for j in range(npt)])
            ptx.append(pt)
            ridx.append(ele + nele)
            nele = nele + ir.GetNPoints()
            ifaces.append(iface)
            self.gtypes[k] = gtype

            self.elattr1[k] = getattr1(i)
            self.elattr2[k] = getattr2(i)

        self.ptx = np.vstack(ptx)
        self.ridx = np.vstack(ridx)
        self.ifaces = np.hstack(ifaces)

        self.emesh_idx = emesh_idx

    def eval(self, expr, solvars, phys, **kwargs):
        refine = kwargs.pop("refine", 1)
        exprs = kwargs.pop("exprs", [expr])
        emesh_idx = get_emesh_idx(self, exprs, solvars, phys)

        if len(emesh_idx) > 1:
            assert False, "expression involves multiple mesh (emesh length != 1)"
        if len(emesh_idx) < 1:
            # use default mesh in this case. (for non mesh dependent expression)
            emesh_idx = [0]
        #    assert False, "expression is not defined on any mesh"

        if (refine != self.refine or self.emesh_idx != emesh_idx[0]):
            self.refine = refine
            self.preprocess_geometry(self.battrs, emesh_idx=emesh_idx[0])
        val = eval_on_edges(self, expr, solvars, phys)
        if val is None:
            return None, None, None

        export_type = kwargs.pop('export_type', 1)

        # print self.ptx.shape, val.shape, self.ridx.shape
        if export_type == 2:
            return self.ptx, val, None
        else:
            return self.ptx, val, self.ridx




'''

 Model tree using OrderedDict.

 This is meant to be a generic layer

'''
from abc import ABC
from collections import OrderedDict
import traceback
from collections.abc import MutableMapping
import six
import os
import numpy as np
import weakref
from weakref import WeakKeyDictionary

from functools import reduce

from petram.namespace_mixin import NS_mixin, NSRef_mixin


def validate_sel(value, obj, w):
    g = obj._global_ns
    l = {"remaining": [], "all": []}
    try:
        value = eval(value, g, l)
        return True
    except:
        return False


def validate_sel2(value, obj, w):
    g = obj._global_ns
    l = {"remaining": [], "all": [], "internal_bdr": []}
    try:
        value = eval(value, g, l)
        return True
    except:
        return False


def convert_sel_txt(txt, g):
    if txt.strip() == 'remaining':
        return ['remaining']
    elif txt.strip() == 'all':
        return ['all']
    elif txt.strip() == 'internal_bdr':
        return ['internal_bdr']
    elif txt.strip() == '':
        arr = []
    else:
        arr = list(np.atleast_1d(eval(txt, g)))
    return arr


class Restorable(object):
    def __init__(self):
        self._requires_restoration = False

    def __setstate__(self, state):
        self.__dict__ = state
        self._requires_restoration = True

    def __getattribute__(self, item):
        if item == '_contents':
            if self._requires_restoration:
                self._requires_restoration = False
                self._restore(self._restoration_data)
                self._restoration_data = None
        return object.__getattribute__(self, item)

    def _restore(self, restoration_data):
        raise NotImplementedError(
            "you must specify the _restore method with the Restorable type")


class RestorableOrderedDict(ABC, MutableMapping, Restorable, object):
    def __init__(self, *args, **kwargs):
        self._contents = OrderedDict(*args, **kwargs)
        Restorable.__init__(self)

    def __setstate__(self, state):
        try:
            Restorable.__setstate__(self, {
                '_contents': OrderedDict(),
                '_restoration_data': state[0],
            })
            self._parent = None
            for attr, value in state[1]:
                setattr(self, attr, value)
        except:
            traceback.print_exc()

    def __getstate__(self):
        st = [(x, self.__dict__[x])
              for x in self.__dict__ if not x.startswith('_')]

        # Note:
        #   we save _sel_index so that model_proc.pmfm has this
        #   informaiton, this way, evaluator does not need to re-run
        #   assign_sel_index, which requires loading base mesh (which is slow)

        if hasattr(self, '_sel_index'):
            st.append(('_sel_index', self._sel_index))

        return [(key, value) for key, value in self._contents.items()], st

    def _restore(self, restoration_data):
        for (key, value) in restoration_data:
            self._contents[key] = value
            value._parent = self

    def __getitem__(self, item):
        if (isinstance(item, list) or
                isinstance(item, tuple)):
            keys = [self]+list(item)
            return reduce(lambda x, y: x[y], keys)
        elif item is None:
            raise KeyError
        elif item.find('.') != -1:
            items = item.split('.')
            keys = [self]+list(items)
            return reduce(lambda x, y: x[y], keys)
        else:
            return self._contents[item]

    def __setitem__(self, key, value):
        self._contents[key] = value
        value._parent = self

    def __delitem__(self, key):
        del self._contents[key]

    def __iter__(self):
        return iter(self._contents)

    def __len__(self):
        return len(self._contents)

    def __repr__(self):
        return """RestorableOrderedDict{}""".format(repr(self._contents))


class Hook(object):
    def __init__(self, names):
        self.names = names


class ModelDict(WeakKeyDictionary):
    '''
    Weak reference dictionary using Model Object (unhashable) as key
    '''

    def __init__(self, root):
        WeakKeyDictionary.__init__(self)
        self.root = weakref.ref(root)

    def __setitem__(self, key, value):
        hook = key.get_hook()
        return WeakKeyDictionary.__setitem__(self, hook, value)

    def __getitem__(self, key):
        hook = key.get_hook()
        return WeakKeyDictionary.__getitem__(self, hook)

    def __iter__(self):
        return [reduce(lambda x, y: x[y], [self.root()] + hook().names)
                for hook in self]


class Model(RestorableOrderedDict):
    can_delete = True
    has_2nd_panel = True
    has_3rd_panel = False
    _has_4th_panel = False
    mustbe_firstchild = False
    always_new_panel = True
    can_rename = False
    unique_child = False
    extra_diagnostic_print = False

    @classmethod
    def fancy_menu_name(cls):
        return cls.__name__.split('_')[-1]

    @classmethod
    def fancy_tree_name(cls):
        return cls.fancy_menu_name()

    def __init__(self, **kwargs):
        super(Model, self).__init__()
        self._parent = None
        self._hook = None

        if not hasattr(self, 'init_attr'):
            self.init_attr = True
            self.update_attribute_set(kw=kwargs)

    @property
    def has_4th_panel(self):
        return self.has_3rd_panel and self._has_4th_panel

    def get_info_str(self):
        return ""

    def get_hook(self):
        if not hasattr(self, '_hook'):
            self._hook = None
        if self._hook is not None:
            return self._hook
        olist = [self]
        o = self
        while o._parent is not None:
            o = o._parent
            olist.append(o)
        names = [o.name() for o in olist]
        self._hook = Hook(names)
        return self._hook

    def __repr__(self):
        return self.__class__.__name__+'('+self.name()+':'+','.join(list(self)) + ')'

    def __eq__(self, x):
        try:
            return x.fullpath() == self.fullpath()
        except:
            return False

    def attribute_set(self, v):
        v['enabled'] = True
        v['sel_readonly'] = False
        v['sel_index'] = []
        v['_sel_index'] = []
        if (hasattr(self, 'sel_index') and
                not hasattr(self, 'sel_index_txt')):
            v['sel_index_txt'] = ', '.join([str(x) for x in self.sel_index])
        elif not hasattr(self, 'sel_index_txt'):
            v['sel_index_txt'] = ''
        else:
            v['sel_index_txt'] = ''
        return v

    def process_sel_index(self, choice=None, internal_bdr=None):
        try:
            arr = convert_sel_txt(self.sel_index_txt, self._global_ns)
            self.sel_index = arr
        except:
            assert False, "failed to convert "+self.sel_index_txt

        if len(self.sel_index) == 1 and self.sel_index[0] == 'remaining':
            self._sel_index = []
            return None
        elif len(self.sel_index) == 1 and self.sel_index[0] == 'all':
            self._sel_index = list(choice)
            return -1
        elif len(self.sel_index) == 1 and self.sel_index[0] == 'internal_bdr':
            if internal_bdr is not None:
                self._sel_index = internal_bdr
            else:
                assert False, "Internal bdr is not defined"
        elif len(self.sel_index) == 0:
            self._sel_index = []
        elif self.sel_index[0] == '':
            self._sel_index = []
        else:
            self._sel_index = [int(i) for i in self.sel_index]
        if choice is not None:
            ret = np.array(self._sel_index)
            ret = list(ret[np.in1d(ret, choice)])
            self._sel_index = ret

        return self._sel_index

    def update_attribute_set(self, kw=None):
        if kw is None:
            kw = {}
        d = self.attribute_set(kw)
        self.do_update_attribute_set(d)

    def do_update_attribute_set(self, d):
        for k in d:
            if not hasattr(self, k):
                try:
                    setattr(self, k, d[k])
                except AttributeError:
                    print("Attribute Error", self, k, d[k])

    def attribute(self, *args, **kwargs):
        if 'showall' in kwargs:
            return {x: self.attribute(x) for x in self.attribute_set({}).keys()}

        if len(args) == 0:
            return list(self.attribute_set({}).keys())
        elif len(args) == 1:
            if hasattr(self, args[0]):
                return getattr(self, args[0])
        elif len(args) == 2:
            if hasattr(self, args[0]):
                setattr(self, args[0], args[1])
        else:
            pass

    def get_editor_menus(self):
        '''
        show custom menu in editor panel
        '''
        return []

    def __setstate__(self, state):
        super(Model, self).__setstate__(state)
        self.update_attribute_set()

    def is_enabled(self):
        ''' 
        check if all parents are all enabled
        '''
        p = self
        while p is not None:
            if not p.enabled:
                return False
            p = p._parent
        return True

    def gather_enebled_flags(self, parent):
        enabled = {}
        for o in parent.walk():
            fname = o.fullname()
            enabled[fname] = o.enabled
        return enabled

    def apply_enebled_flags(self, parent, enabled):
        for o in parent.walk():
            fname = o.fullname()
            if fname in enabled:
                o.enabled = enabled[fname]

    @property
    def parent(self):
        return self._parent

    @property
    def parents(self):
        parents = []
        p = self
        while True:
            p = p._parent
            if p is None:
                break
            parents.insert(0, p)
        return parents

    def verify_setting(self):
        '''
        a check routine to seting verificaiton comes here.
        return flag,  text, long explanation
        '''
        return True, '', ''

    def GetItemText(self, indices):
        key = ''
        d0 = self
        for k in indices:
            key = list(d0.keys())[k]
            d0 = d0[key]
        return key

    def GetChildrenCount(self, indices):
        d0 = self
        for k in indices:
            key = list(d0.keys())[k]
            d0 = d0[key]
        return len(d0)

    def GetItem(self, indices):
        d0 = self
        for k in indices:
            key = list(d0)[k]
            d0 = d0[key]
        return d0

    def GetIndices(self):
        parents = self.parents+[self, ]
        indices = [list(parents[i]).index(parents[i+1].name())
                   for i in range(len(parents)-1)]
        return indices

    def get_child(self, id):
        return self[list(self.keys())[id]]

    def get_children(self):
        return list(self.values())

    def get_possible_child(self):
        return []

    def get_possible_child_menu(self):
        return [('', cls) for cls in self.get_possible_child()]

    def get_special_menu(self, evt):
        return []

    def add_item(self, txt, cls,  **kwargs):
        after = kwargs.pop("after", None)
        before = kwargs.pop("before", None)

        if cls.unique_child:
            if txt in self:
                assert False, "this class (unique_child) already exists"
            obj = cls(**kwargs)
            self[txt] = obj
            return txt

        m = []
        for k in self.keys():
            ll = len(k)
            while ll >= 0:
                if k[ll-1].isdigit():
                    ll = ll-1
                else:
                    break

            name = k[:ll]
            if name == txt:
                if len(k) > len(name):
                    m.append(int(k[ll:]))

        if len(m) == 0:
            name = txt+str(1)
        else:
            name = txt + str(max(m)+1)

        obj = cls(**kwargs)
        done = False
        if obj.mustbe_firstchild:
            old_contents = self._contents
            self._contents = OrderedDict()
            self[name] = obj
            names = list(old_contents)
            for n in names:
                self[n] = old_contents[n]
            done = True
        elif after is not None:
            old_contents = self._contents
            self._contents = OrderedDict()
            names = list(old_contents)
            for n in names:
                self[n] = old_contents[n]
                if n == after.name():
                    self[name] = obj
                    done = True
                    # break
        elif before is not None:
            old_contents = self._contents
            self._contents = OrderedDict()
            names = list(old_contents)
            for n in names:
                if n == before.name():
                    self[name] = obj
                    done = True
                self[n] = old_contents[n]
                # if done: break

        if not done:
            self[name] = obj
        return name

    def postprocess_after_add(self, engine):
        pass

    def add_itemobj(self, txt, obj, nosuffix=False):

        m = []
        for k in self.keys():
            if k.startswith(txt):
                for x in k[len(txt):]:
                    if not x.isdigit():
                        break
                else:
                    m.append(int(k[len(txt):]))
        if len(m) == 0:
            if nosuffix:
                name = txt
            else:
                name = txt+str(1)
        else:
            name = txt + str(max(m)+1)
        self[name] = obj
        return name

    def panel2_tabname(self):
        return "Selection"

    def panel3_tabname(self):
        return "Init."

    def panel4_tabname(self):
        return "Time Dep./NL."

    def panel1_param(self):
        return []

    def panel2_param(self):
        return []

    def panel2_sel_labels(self):
        return ['selection']

    def panel2_all_sel_index(self):
        try:
            idx = [int(x) for x in self.sel_index]
        except:
            idx = []
        return [idx]

    def is_wildcard_in_sel(self):
        ans = [False, ]
        try:
            idx = [int(x) for x in self.sel_index]
        except:
            ans[0] = True
        return ans

    def panel3_param(self):
        return []

    def panel4_param(self):
        return []

    def panel1_tip(self):
        return None

    def panel2_tip(self):
        return None

    def panel3_tip(self):
        return None

    def panel4_tip(self):
        return None

    def get_panel1_value(self):
        return []

    def get_panel2_value(self):
        return (self.sel_index_txt,)

    def get_panel3_value(self):
        return []

    def get_panel4_value(self):
        return []

    def import_panel1_value(self, v):
        '''
        return value : gui_update_request
        '''
        return False

    def import_panel2_value(self, v):
        '''
        return value : gui_update_request
        '''
        if not self.sel_readonly:
            self.sel_index_txt = str(v[0])
            try:
                arr = convert_sel_txt(str(v[0]), self._global_ns)
                self.sel_index = arr
            except:
                pass
        return False

    def import_panel3_value(self, v):
        '''
        return value : gui_update_request
        '''
        return False

    def import_panel4_value(self, v):
        '''
        return value : gui_update_request
        '''
        return False

    def onItemSelChanged(self, evt):
        '''
        GUI response when model object is selected in
        the dlg_edit_model
        '''
        viewer = evt.GetEventObject().GetTopLevelParent().GetParent()
        viewer.set_view_mode('',  self)

    def export_modeldata(self):
        pass

    def write_setting(self, fid):
        pass

    def preprocess_params(self, engine):
        pass

    def walk(self):
        yield self
        for k in self.keys():
            for x in self[k].walk():
                yield x

    def walk_enabled(self, skip_self=False):
        '''
        skip_self: not return the top level model
        '''
        if not self.enabled:
            return
        if not skip_self:
            yield self
        for k in self.keys():
            if not self[k].enabled:
                continue
            for x in self[k].walk_enabled():
                yield x

    def iter_enabled(self):
        for child in self.values():
            if not child.enabled:
                continue
            yield child
    enum_enabled = iter_enabled  # backward compabibility.

    def name(self):
        if self._parent is None:
            return 'root'
        for k in self._parent.keys():
            if self._parent[k] is self:
                return k
        return "No Parent"

    def rename(self, new_name):
        if self.name() == 'root':
            assert False, "can't rename root"

        new_cnt = []
        for key in self._parent.keys():
            if self._parent[key] is self:
                new_cnt.append((new_name, self._parent[key]))
            else:
                new_cnt.append((key, self._parent[key]))

        parent = self._parent
        for key in list(self._parent):
            parent[key]._parent = None
            del parent[key]

        for key, value in new_cnt:
            parent[key] = value

    def split_digits(self):
        '''
        split tailing digits
        '''
        name = self.name()
        l = -1
        if not name[l].isdigit():
            return name, '0'

        while name[l].isdigit():
            l = l-1
        l = l+1
        return name[:l], name[l:]

    def insert_item(self, index, name, item):
        items = list(self._contents.items())
        items.insert(index, (name, item))
        self._contents = OrderedDict(items)
        item.set_parent(self)

    def set_parent(self, parent):
        self._parent = parent

    def on_created_in_tree(self):
        # called when item is newly created in Tree from GUI
        None

    def fullname(self):
        '''
        returns 'root.Phys.Boundary...'
        '''
        olist = [self]
        o = self
        while o._parent is not None:
            o = o._parent
            olist.append(o)
        names = [o.name() for o in olist]
        return '.'.join(reversed(names))

    def fullpath(self):
        '''
        returns 'Phys.Boundary...'
        similar to fullname but without "root"
        can be used to root[obj.fullpath()] to get model object
        '''
        olist = [self]
        o = self
        while o._parent is not None:
            o = o._parent
            olist.append(o)
        names = [o.name() for o in olist]
        return '.'.join(reversed(names[:-1]))

    def root(self):
        o = self
        while o._parent is not None:
            o = o._parent
        return o

    def has_ns(self):
        return isinstance(self, NS_mixin)

    def has_nsref(self):
        return isinstance(self, NSRef_mixin)

    def add_node(self, name='', cls=''):
        ''' 
        this is similar to add_item, but does not
        do anything to modify name
        '''
        if not name in self.keys():
            self[name] = cls()
        return self[name]

    def set_script_idx(self, idx=1):
        self._script_name = 'obj'+str(idx)

        for name in self.keys():
            node = self[name]
            idx = idx + 1
            idx = node.set_script_idx(idx=idx)
        return idx

    def save_attribute_set(self, skip_def_check):
        ans = []
        for attr in self.attribute():
            defvalue = self.attribute_set(dict())
            value = self.attribute(attr)
            mycheck = True

            try:
                # print attr, type(value), value, type(defvalue[attr]), defvalue[attr]
                if type(value) != type(defvalue[attr]):
                    mycheck = True
                else:  # for numpy array
                    mycheck = value != defvalue[attr]  # for numpy array
                    if isinstance(mycheck, np.ndarray):
                        mycheck = mycheck.any()
                    else:
                        mycheck = any(mycheck)
            except TypeError:
                try:
                    mycheck = value != defvalue[attr]
                except:
                    pass
            if mycheck or skip_def_check:
                ans.append(attr)
        return ans

    def _generate_model_script(self, script=None,
                               skip_def_check=False,
                               dir=None):
        # assigne script index if root node
        if script is None:
            self.set_script_idx()
            script = []
            script.append('obj1 = MFEM_ModelRoot()')

        attrs = self.save_attribute_set(skip_def_check)
        for attr in attrs:
            value = self.attribute(attr)
            script.append(self._script_name + '.'+attr + ' = ' +
                          value.__repr__())

        if (self.has_ns() or self.has_nsref()) and self.ns_name is not None:
            script.append(self._script_name + '.ns_name = "' +
                          self.ns_name + '"')

        ns_names = []
        for name in self.keys():
            node = self[name]

            script.append(node._script_name +
                          ' = ' + self._script_name +
                          '.add_node(name = "'+name + '"' +
                          ', cls = '+node.__class__.__name__ + ')')
            script = node._generate_model_script(script=script,
                                                 skip_def_check=skip_def_check,
                                                 dir=dir)
            if node.has_ns():
                if node.ns_name is None:
                    continue
                if not node.ns_name in ns_names:
                    ns_names.append(node.ns_name)
                    node.write_ns_script_data(dir=dir)

        return script

    def generate_main_script(self):

        script = []
        script.append('import time, datetime')
        script.append('stime = time.time()')
        script.append('if mfem_config.use_parallel:')
        script.append('    from petram.engine import ParallelEngine as Eng')
        script.append('else:')
        script.append('    from petram.engine import SerialEngine as Eng')
        script.append('')
        script.append('import petram.debug as debug')
        script.append('debug.set_debug_level(debug_level)')
        script.append('')
        script.append('model = make_model()')
        script.append('')
        script.append('eng = Eng(model = model)')
        script.append('')

        script.append('solvers = eng.run_build_ns()')
        # script.append('solvers = eng.preprocess_modeldata()')
        # script.append('if myid == 0: model.save_to_file("model_proc.pmfm", meshfile_relativepath = False)')
        script.append('')
        script.append('is_first = True')
        script.append('for s in solvers:')
        script.append('    s.run(eng, is_first=is_first)')
        script.append('    is_first=False')
        script.append('')
        script.append('if myid == 0:')
        script.append('    print("End Time " + ')
        script.append(
            '          datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S_%f"))')
        script.append(
            '    print("Total Elapsed Time: " + str(time.time()-stime) + "s")')
        script.append('    print("Petra-M Normal End")')

        return script

    def generate_import_section(self, script):
        from petram.helper.variables import var_g
        from collections import OrderedDict

        d1 = OrderedDict()
        d2 = OrderedDict()
        for x in var_g.keys():
            err = False
            try:
                d1[x] = var_g[x].__module__
            except:
                try:
                    d1[x] = var_g[x].__class__.__module__
                except:
                    assert False, "can not generate script for " + x

        for x in self.walk():
            d2[x.__class__.__name__] = x.__class__.__module__

        for key in d1.keys():
            if d1[key] == '__builtin__':
                continue  # skip something like pi....(PY2)
            if d1[key] == 'builtins':
                continue  # skip something like pi....(PY3)
            script.append('from '+d1[key] + ' import ' + key)
        script.append('')
        for key in d2.keys():
            script.append('from '+d2[key] + ' import ' + key)

        script.append('from collections import OrderedDict')

    def generate_script(self, skip_def_check=False, dir=None, nofile=False,
                        parallel=False, filename='model.py'):
        if dir is None:
            dir = os.getcwd()
        script = []
        script.extend(['from __future__ import print_function',
                       'import os',
                       '',
                       'if __name__=="__main__":',
                       '    from mfem.common.arg_parser import ArgParser',
                       '    parser = ArgParser(description="PetraM sciprt")',
                       '    parser.add_argument("-s", "--force-serial", ',
                       '                     action = "store_true", ',
                       '                     default = True,',
                       '                     help="Use serial model even if nproc > 1.")',
                       '    parser.add_argument("-p", "--force-parallel", ',
                       '                     action = "store_true", ',
                       '                     default = False,',
                       '                     help="Use parallel model even if nproc = 1.")',
                       '    parser.add_argument("-d", "--debug-param", ',
                       '                     action = "store", ',
                       '                     default = 1, type=int) ',
                       '',
                       '    args = parser.parse_args()',
                       '    if args.force_parallel:',
                       '        use_parallel = True',
                       '        args.force_serial = False',
                       '    else:',
                       '        use_parallel = False',
                       '',
                       '    import  petram.mfem_config as mfem_config',
                       '    mfem_config.use_parallel = use_parallel',
                       '    debug_level=args.debug_param',
                       '',
                       '# this is needed if this file is being imported',
                       'if not "use_parallel" in locals():',
                       '    use_parallel = False',
                       '#set default parallel/serial flag',
                       'if use_parallel:',
                       '    from mpi4py import MPI',
                       '    num_proc = MPI.COMM_WORLD.size',
                       '    myid = MPI.COMM_WORLD.rank',
                       'else:',
                       '    myid = 0',
                       '    num_proc = 1', ])

        script.append('')
        self.generate_import_section(script)
        script.append('')

        script2 = self._generate_model_script(
            skip_def_check=skip_def_check,
            dir=dir)

        script.append('def make_model():')
        for x in script2:
            script.append(' '*4 + x)
        script.append(' '*4 + 'return obj1')

        script.append('')
        script.extend(['if __name__ == "__main__":',
                       '    if (myid == 0): parser.print_options(args)'
                       '', ])

        script.append('')
        main_script = self.generate_main_script()
        for x in main_script:
            script.append(' '*4 + x)

        path1 = os.path.join(dir, filename)
        fid = open(path1, 'w')
        fid.write('\n'.join(script))
        fid.close()

        return script

    def load_gui_figure_data(self, viewer):
        '''
        called when mfem_viewer opened to set inital figure (geometry)
        plottting data.

        return value : (view_mode, name, data)
        '''
        return None, None, None

    def is_viewmode_grouphead(self):
        return False

    def update_figure_data(self, *args, **kwargs):
        del args
        del kwargs

    def figure_data_name(self):
        return self.name()

    def update_after_ELChanged2(self, evt):
        '''
        return True if edit panel needs update
        '''
        return False

    def update_after_ELChanged(self, dlg):
        '''
        return True if edit panel needs update
        '''
        return False

    def use_essential_elimination(self):
        return True


class Bdry(Model):
    can_delete = True
    is_essential = False

    def attribute_set(self, v):
        v = super(Bdry, self).attribute_set(v)
        v['sel_readonly'] = True
        v['esse_elim_txt'] = True

        return v

    def get_possible_child(self):
        return self.parent.get_possible_bdry()

    def panel2_param(self):
        return [["Boundary",  'remaining',  0, {'changing_event': True,
                                                'setfocus_event': True,
                                                'validator': validate_sel2,
                                                'validator_param': self}]]

    def use_essential_elimination(self):
        return self.esse_elim_txt


class Pair(Model):
    can_delete = True
    is_essential = False

    def attribute_set(self, v):
        v = super(Pair, self).attribute_set(v)
        v['sel_readonly'] = False
        v['src_index'] = []
        v['sel_index'] = []
        v['src_index_txt'] = ''
        v['sel_index_txt'] = ''

        return v

    def get_possible_child(self):
        return self.parent.get_possible_pair()

    def panel2_param(self):
        return [["Source",  '',  0, {'changing_event': True,
                                     'setfocus_event': True,
                                     'validator': validate_sel,
                                     'validator_param': self}],
                ["Destination",  '',  0, {'changing_event': True,
                                          'setfocus_event': True,
                                          'validator': validate_sel,
                                          'validator_param': self}], ]

    def panel2_sel_labels(self):
        return ['source', 'destination']

    def panel2_all_sel_index(self):
        try:
            idx = [int(x) for x in self.sel_index]
        except:
            idx = []
        try:
            idx2 = [int(x) for x in self.src_index]
        except:
            idx2 = []

        return [idx2, idx]

    def is_wildcard_in_sel(self):
        ans = [False, False]
        try:
            idx = [int(x) for x in self.sel_index]
        except:
            ans[0] = True
        try:
            idx2 = [int(x) for x in self.src_index]
        except:
            ans[1] = True
        return ans

    def get_panel2_value(self):
        return (self.src_index_txt, self.sel_index_txt)

    def import_panel2_value(self, v):

        self.src_index_txt = str(v[0])
        self.sel_index_txt = str(v[1])

        try:
            arr = convert_sel_txt(str(v[0]), self._global_ns)
            self.src_index = arr
        except:
            pass
        try:
            arr = convert_sel_txt(str(v[1]), self._global_ns)
            self.sel_index = arr
        except:
            pass
        return False

    def process_sel_index(self, choice=None, internal_bdr=None):
        # interanl_bdr is ignored..
        try:
            arr = convert_sel_txt(self.src_index_txt, self._global_ns)
            self.src_index = arr
        except:
            assert False, "failed to convert "+self.src_index_txt

        try:
            arr = convert_sel_txt(self.sel_index_txt, self._global_ns)
            self.sel_index = arr
        except:
            assert False, "failed to convert "+self.sel_index_txt

        if len(self.sel_index) == 0:
            self._sel_index = []
        elif self.sel_index[0] == '':
            self._sel_index = []
        else:
            self._sel_index = [int(i) for i in self.sel_index]
        if len(self.src_index) == 0:
            self._src_index = []
        elif self.src_index[0] == '':
            self._src_index = []
        else:
            self._src_index = [int(i) for i in self.src_index]
        self._sel_index = self._sel_index + self._src_index

        if choice is not None:
            ret = np.array(self._sel_index)
            ret = list(ret[np.in1d(ret, choice)])
            self._sel_index = ret
            ret = np.array(self._src_index)
            ret = list(ret[np.in1d(ret, choice)])
            self._src_index = ret

        return self._sel_index


class Domain(Model):
    can_delete = True
    is_essential = False

    def attribute_set(self, v):
        v = super(Domain, self).attribute_set(v)
        v['sel_readonly'] = True
        return v

    def get_possible_child(self):
        return self.parent.get_possible_domain()

    def panel2_param(self):
        return [["Domain",  'remaining',  0, {'changing_event': True,
                                              'setfocus_event': True,
                                              'validator': validate_sel,
                                              'validator_param': self}], ]


class Edge(Model):
    can_delete = True
    is_essential = False

    def attribute_set(self, v):
        v = super(Edge, self).attribute_set(v)
        v['sel_readonly'] = True
        return v

    def get_possible_child(self):
        return self.parent.get_possible_edge()

    def panel2_param(self):
        return [["Edge",  'remaining',  0, {'changing_event': True,
                                            'setfocus_event': True,
                                            'validator': validate_sel,
                                            'validator_param': self}], ]


class Point(Model):
    can_delete = True
    is_essential = False

    def attribute_set(self, v):
        v = super(Point, self).attribute_set(v)
        v['sel_readonly'] = True
        return v

    def get_possible_child(self):
        return self.parent.get_possible_point()

    def panel2_param(self):
        return [["Point",  'remaining',  0, {'changing_event': True,
                                             'setfocus_event': True,
                                             'validator': validate_sel,
                                             'validator_param': self}], ]




'''
    read .nas file and make MFEM mesh file

    note:
       it attempts to eliminate low order surfaces.
       it supports 2nd order element only for a pure tet mesh.
    supported card
        GRID
        CTETRA
        CHEXA
        CTRIA6
        CTRIA3
        CQUAD8
        PSOLID
        PSHELL

    index in nas starts from 1.
    index in mfem starts from 0.

'''
import numba
from numba import jit, types, prange
import numpy as np
import re
from scipy.sparse import coo_matrix


fwidth = 16


class NASReader(object):
    def __init__(self, filename):
        self.filename = filename
        self.dataset = None

    def load(self):
        # check format length
        num_lines = sum(1 for line in open(self.filename))
        num_lines_base = num_lines/20
        print("number of lines in file: ", num_lines)
        fid = open(self.filename, 'r')

        print("checking format...")
        counter = 0
        counter2 = 0
        while True:
            l = fid.readline()
            if l == '':
                print("cannot determine file format whether it is small or large format")
                print("processing the file assuming it is a small format file")
                globals()['fwidth'] = 8
                break
            if l.startswith('+CONT'):
                globals()['fwidth'] = 8
                print('short format ' + str(num_lines) + ' lines')
                break
            elif l.startswith('*CONT'):
                globals()['fwidth'] = 16
                print('long format ' + str(num_lines) + ' lines')
                break
            counter = counter + 1
            if int(counter % (num_lines/5)) == 0:
                print("read ", (counter2*20), "% done")
                counter2 = counter2 + 1
        fid.close()

        print("...done")

        fid = open(self.filename, 'r')
        while True:
            l = fid.readline()
            if l.startswith('$ Grid data section'):
                break

        num_grid = 0
        while True:
            l = fid.readline()
            if l.startswith('GRID'):
                num_grid = num_grid + 1
            if l.startswith('$ Element data section'):
                break
        fid.close()
        fid = open(self.filename, 'r')
        cl = ''

        while True:
            l = self._read_line(fid)
            if l.startswith('$ Grid data section'):
                break

        grids = [None]*num_grid  # reading grid
        for i in range(num_grid):
            l = self._read_line(fid)
            if l.startswith('$ Element data section'):
                break
            g = self.parse_grid_fixed(l)
            grids[i] = np.array((float(g[3]), float(g[4]), float(g[5])))
        grids = np.vstack(grids)

        elems = {'TRIA6': [],
                 'TRIA3': [],
                 'TETRA': [],
                 'HEXA': [],
                 'QUAD4': [],
                 'QUAD8': []}  # reading elements
        print("reading elements")
        ll = 0
        while True:
            l = self._read_line(fid)
            if l.startswith('$ Property data section'):
                break
            if l.startswith('$ Element data section'):
                continue
            if l.startswith('CTRIA6'):
                elems['TRIA6'].append(self.parse_tria6_fixed(l))
            elif l.startswith('CTRIA3'):
                elems['TRIA3'].append(self.parse_tria3_fixed(l))
            elif l.startswith('CTETRA'):
                elems['TETRA'].append(self.parse_tetra_fixed(l))
            elif l.startswith('CHEXA'):
                elems['HEXA'].append(self.parse_hexa_fixed(l))
            elif l.startswith('CQUAD8'):
                elems['QUAD8'].append(self.parse_quad8_fixed(l))
            elif l.startswith('CQUAD4'):
                elems['QUAD4'].append(self.parse_quad4_fixed(l))
            else:
                print("Element not supported: " + l)
                continue
            ll += 1
            if int(ll % num_lines_base) == 0:
                print(str(int(ll/num_lines_base)*5) +
                      "% done.(" + str(ll) + ")")
        print("reading elements   (done)")
        for key in elems:
            if len(elems[key]) > 0:
                print(str(len(elems[key])) + " elements of " + key)
        new_elems = {}
        new_elems_f = {}

        edge_matrix_3d = coo_matrix((len(grids), len(grids)), dtype=bool)
        has_ho = False

        is_3d = len(elems['TETRA']) > 0 or len(elems['HEXA']) > 0

        if len(elems['TETRA']) > 0:
            print("Processing TETRA...")
            TETRA = np.vstack([np.array([int(x) for x in g[3:7]])
                               for g in elems['TETRA']])
            # [np.array((int(g[3]), int(g[4]), int(g[5]), int(g[6]),))
            #       for g in elems['TETRA']])
            TETRAF = np.vstack([np.array([int(x) for x in g[3:]])
                                for g in elems['TETRA']])
            TETRA_ATTR = np.array([int(g[2])
                                   for g in elems['TETRA']])  # PSOLID ID
            idx = [len(np.unique(x)) == 4 for x in TETRA]
            if not all(idx):
                print("some TETRA has no volume")
                TETRA = TETRA[idx, :]
                TETRA_ATTR = TETRA_ATTR[idx]

            new_elems['TETRA'] = TETRA-1
            new_elems['TETRA_ATTR'] = TETRA_ATTR
            new_elems_f['TETRA'] = TETRAF-1

            idx = [(0, 1), (1, 2), (2, 0), (3, 0),
                   (3, 1), (3, 2)]

            d1 = np.hstack([new_elems['TETRA'][:, x] for x, y in idx])
            d2 = np.hstack([new_elems['TETRA'][:, y] for x, y in idx])
            d3 = np.ones(len(d2), dtype=bool)

            edge_rows_tet = np.max(np.vstack((d1, d2)), axis=0)
            edge_cols_tet = np.min(np.vstack((d1, d2)), axis=0)
            # edge_rows_tet = np.hstack([d1, d2])#max(np.vstack((d1, d2)), axis=0)
            # edge_cols_tet = np.hstack([d2, d1])#np.min(np.vstack((d1, d2)), axis=0)

            edge_matrix_3d.row = np.hstack((edge_matrix_3d.row, edge_rows_tet))
            edge_matrix_3d.col = np.hstack((edge_matrix_3d.col, edge_cols_tet))
            edge_matrix_3d.data = np.hstack((edge_matrix_3d.data, d3))
            print("done")

        if len(elems['HEXA']) > 0:
            print("Processing HEXA...")
            HEXA = np.vstack([np.array((int(g[3]), int(g[4]), int(g[5]), int(g[6]),
                                        int(g[7]), int(g[8]), int(g[9]), int(g[10]),))
                              for g in elems['HEXA']])
            TEXAF = np.vstack([np.array([int(x) for x in g[3:]])
                               for g in elems['HEXA']])
            HEXA_ATTR = np.array([int(g[2])
                                  for g in elems['HEXA']])  # PSOLID ID
            new_elems['HEXA'] = HEXA-1
            new_elems['HEXA_ATTR'] = HEXA_ATTR
            new_elems_f['HEXA'] = HEXAF-1

            idx = [(0, 1), (1, 2), (2, 3), (3, 0),
                   (0, 4), (1, 5), (2, 6), (3, 7),
                   (4, 5), (5, 6), (6, 7), (7, 0), ]

            d1 = np.hstack([new_elems['HEXA'][:, x] for x, y in idx])
            d2 = np.hstack([new_elems['HEXA'][:, y] for x, y in idx])
            d3 = np.ones(len(d2), dtype=bool)

            edge_rows_hex = np.max(np.vstack((d1, d2)), axis=0)
            edge_cols_hex = np.min(np.vstack((d1, d2)), axis=0)
            # edge_rows_tet = np.hstack([d1, d2])#max(np.vstack((d1, d2)), axis=0)
            # edge_cols_tet = np.hstack([d2, d1])#np.min(np.vstack((d1, d2)), axis=0)

            edge_matrix_3d.row = np.hstack((edge_matrix_3d.row, edge_rows_hex))
            edge_matrix_3d.col = np.hstack((edge_matrix_3d.col, edge_cols_hex))
            edge_matrix_3d.data = np.hstack((edge_matrix_3d.data, d3))
            print("done")

        edge_matrix_3d = edge_matrix_3d.tocsr()

        if len(elems['TRIA6']) > 0:
            print("Processing TRIA6...")
            TRIA6 = np.vstack([np.array((int(g[3]), int(g[4]), int(g[5]), ))
                               for g in elems['TRIA6']])
            TRIA6F = np.vstack([np.array([int(x) for x in g[3:]])
                                for g in elems['TRIA6']])
            TRIA6_ATTR = np.array([int(g[2]) for g in elems['TRIA6']])

            idx = [len(np.unique(x)) == 3 for x in TRIA6]
            if not all(idx):
                print("some TRIA6 has no area")
                TRIA6 = TRIA6[idx, :]
                TRIA6F = TRIA6F[idx, :]
                TRIA6_ATTR = TRIA6_ATTR[idx]

            idx = [(0, 1), (1, 2), (2, 0)]

            TRIA6 = TRIA6 - 1

            d1 = np.hstack([TRIA6[:, x] for x, y in idx])
            d2 = np.hstack([TRIA6[:, y] for x, y in idx])
            rows = np.max(np.vstack((d1, d2)), axis=0)
            cols = np.min(np.vstack((d1, d2)), axis=0)

            flags = np.array(edge_matrix_3d[rows, cols].reshape(3, -1))
            flags = (np.sum(flags, axis=0) == 3).flatten()

            if np.sum(flags) != flags.shape[0]:
                print("some TRIA6 is not surface of TETRA!",
                      flags.shape[0]-np.sum(flags))
                # save the location of strange tria6
                self.garbage = TRIA6[np.logical_not(flags), :]
                TRIA6 = TRIA6[flags, :]
                TRIA6_ATTR = TRIA6_ATTR[flags]

            self.edge_matrix_3d = edge_matrix_3d

            new_elems['TRIA6'] = TRIA6
            new_elems['TRIA6_ATTR'] = TRIA6_ATTR
            new_elems_f['TRIA6'] = TRIA6F-1

            if is_3d:
                if len(elems['QUAD8']) > 0:
                    print(
                        "!!!! this 3D mesh file contains ho quad. this is not supported.")
                    print("!!! linear mesh will be generated")
                else:
                    has_ho = True
            print("done")

        if len(elems['TRIA3']) > 0:
            TRIA3 = np.vstack([np.array((int(g[3]), int(g[4]), int(g[5]), ))
                               for g in elems['TRIA3']])
            TRIA3_ATTR = np.array([int(g[2])
                                   for g in elems['TRIA3']])  # PSHELL ID
            new_elems['TRIA3'] = TRIA3-1
            new_elems['TRIA3_ATTR'] = TRIA3_ATTR
            new_elems_f['TRIA3'] = TRIA3-1

        if len(elems['QUAD8']) > 0:
            print("Processing QUAD8...")
            QUAD8 = np.vstack([np.array((int(g[3]), int(g[4]), int(g[5]), int(g[6])))
                               for g in elems['QUAD8']])
            QUAD8F = np.vstack([np.array([int(x) for x in g[3:]])
                                for g in elems['QUAD8']])
            QUAD8_ATTR = np.array([int(g[2])
                                   for g in elems['QUAD8']])  # PSHELL ID
            new_elems['QUAD8'] = QUAD8-1
            new_elems['QUAD8_ATTR'] = QUAD8_ATTR
            new_elems_f['QUAD8'] = QUAD8F-1

            print("done")

        if len(elems['QUAD4']) > 0:
            QUAD4 = np.vstack([np.array((int(g[3]), int(g[4]), int(g[5]), int(g[6])))
                               for g in elems['QUAD4']])
            QUAD4_ATTR = np.array([int(g[2])
                                   for g in elems['QUAD4']])  # PSHELL ID
            new_elems['QUAD4'] = QUAD4-1
            new_elems['QUAD4_ATTR'] = QUAD4_ATTR
            new_elems_f['QUAD4'] = QUAD4-1

        elems = new_elems

        props = {'PSOLID': [],
                 'PSHELL': []}
        print("reading shell/solid")
        while True:
            l = self._read_line(fid)
            if l.startswith('ENDDATA'):
                break
            if l.startswith('PSOLID'):
                props['PSOLID'].append(self.parse_psolid_fixed(l))
            elif l.startswith('PSHELL'):
                props['PSHELL'].append(self.parse_pshell_fixed(l))
            else:
                pass
            ll += 1
            if ll % num_lines_base == 0:
                print(str(ll/num_lines_base) + "% done.")
        print("reading shell/solid ...(done)")
        PSHELL = np.array([int(g[1]) for g in props['PSHELL']])  # PSHELL
        PSOLID = np.array([int(g[1]) for g in props['PSOLID']])  # PSOLID

        props = {'PSOLID': PSOLID,
                 'PSHELL': PSHELL}

        dataset = {'PROPS': props,
                   'ELEMS': elems,
                   'GRIDS': grids}
        if has_ho:
            dataset["ELEMS_HO"] = new_elems_f
            self.has_ho = True
        else:
            self.has_ho = False

        fid.close()
        self.dataset = dataset

    def plot_tet(self, idx, **kwargs):
        from ifigure.interactive import solid

        grids = self.dataset['GRIDS']
        tet = self.dataset['ELEMS']['TETRA']
        i = tet[idx]
        pts = [grids[[i[0], i[1], i[2]]],
               grids[[i[1], i[2], i[3]]],
               grids[[i[2], i[3], i[0]]],
               grids[[i[3], i[0], i[1]]]]
        pts = np.rollaxis(np.dstack(pts), 2, 0)
        solid(pts, **kwargs)

    def _read_line(self, fid):
        cl = ''
        while True:
            line = fid.readline()
            l = line.rstrip("\r\n")
            if (l.startswith('+CONT') or
                    l.startswith('*CONT')):
                l = cl + l[8:]
            if (l.strip().endswith('+CONT') or
                    l.strip().endswith('*CONT')):
                cl = l.strip()[:-5]
                continue
            break
        return l

    def parse_grid_fixed(self, l):
        d = fwidth
        if fwidth == 16:
            l = ' '*8+l
#        cards= [l[d*i:d*(i+1)].strip() for i in range(8)]
        cards = re.findall('.'*d, l)
        return cards

    def parse_tetra_fixed(self, l):
        d = 8
#        cards= [l[d*i:d*(i+1)].strip() for i in range(13)]
        cards = re.findall('.'*d, l)
        return cards

    def parse_tria6_fixed(self, l):
        d = 8
#        cards= [l[d*i:d*(i+1)].strip() for i in range(9)]
        cards = re.findall('.'*d, l)
        return cards

    def parse_tria3_fixed(self, l):
        d = 8
#        cards= [l[d*i:d*(i+1)].strip() for i in range(9)]
        cards = re.findall('.'*d, l)
        return cards

    def parse_hexa_fixed(self, l):
        d = 8
#        cards= [l[d*i:d*(i+1)].strip() for i in range(23)]
        cards = re.findall('.'*d, l)
        return cards

    def parse_quad8_fixed(self, l):
        d = 8
#        cards= [l[d*i:d*(i+1)].strip() for i in range(11)]
        cards = re.findall('.'*d, l)
        return cards

    def parse_quad4_fixed(self, l):
        d = 8
#        cards= [l[d*i:d*(i+1)].strip() for i in range(11)]
        cards = re.findall('.'*d, l)
        return cards

    def parse_pshell_fixed(self, l):
        d = 8
#        cards= [l[d*i:d*(i+1)].strip() for i in range(2)]
        cards = re.findall('.'*d, l)
        return cards

    def parse_psolid_fixed(self, l):
        d = 8
        cards = re.findall('.'*d, l)
#        cards= [l[d*i:d*(i+1)].strip() for i in range(2)]
        return cards


def dist(p1, p2):
    d = np.sqrt(np.sum(p1 - p2)**2)
    return d


def linearized_grids(grids):
    new_grids = np.vstack([grids[0], grids[1],  grids[2], grids[3],
                           (grids[0] + grids[1])/2.0,
                           (grids[1] + grids[2])/2.0,
                           (grids[0] + grids[2])/2.0,
                           (grids[0] + grids[3])/2.0,
                           (grids[1] + grids[3])/2.0,
                           (grids[2] + grids[3])/2.0])
    return new_grids


def prepare_ho_node(mesh):
    import mfem.ser as mfem
    nfec = mfem.H1_FECollection(2,
                                mesh.Dimension())
    # mfem.BasisType.ClosedUniform)
    nfes = mfem.FiniteElementSpace(mesh,
                                   nfec,
                                   mesh.SpaceDimension(),
                                   mfem.Ordering.byVDIM)
    mesh.SetNodalFESpace(nfes)
    mesh._linked_obj = (nfes, nfec)


def write_ho_tet_mesh(filename, reader, ho_thre=1e-20):
    import mfem.ser as mfem

    @jit(numba.int64(numba.float64[:], numba.float64[:, :]), cache=False)
    def find_closest0(ptx, grids):
        x1 = ptx[0]
        y1 = ptx[1]
        z1 = ptx[2]
        for jj in range(10):
            d2 = ((x1 - grids[jj, 0])**2 +
                  (y1 - grids[jj, 1])**2 +
                  (z1 - grids[jj, 2])**2)
            if d2 < ho_thre:
                return jj
        assert(False)
        return -1

    @jit(parallel=True, cache=False)
    def find_closest_tet_ptx(ptx, grids):
        idx = np.empty(10, dtype=numba.int64)
        for jj in prange(10):
            idx[jj] = find_closest0(ptx[jj], grids)
        return idx

    mesh = mfem.Mesh(filename)

    prepare_ho_node(mesh)
    n = mesh.GetNE()

    node_gf = mesh.GetNodes()
    nfes = node_gf.FESpace()
    nodes = node_gf.GetDataArray().copy()

    node_gf2 = mfem.GridFunction(mesh._linked_obj[0])
    nodes2 = node_gf2.GetDataArray()

    grid_all = reader.dataset['GRIDS']
    i_tetra = reader.dataset["ELEMS_HO"]["TETRA"]

    for i in range(n):
        # for i in [0]:
        vdofs = nfes.GetElementVDofs(i)
        ptx = nodes[vdofs].reshape(3, -1).transpose()
        grids = grid_all[i_tetra[i]]

        new_grids = linearized_grids(grids)

        idx = find_closest_tet_ptx(ptx, new_grids)

        tmp = grids[idx]
        data = np.hstack((tmp[:, 0], tmp[:, 1], tmp[:, 2]))
        nodes2[vdofs] = data

        if (i % 50000) == 0:
            print("processing elements "+str(i) + "/" + str(n))

    node_gf.Assign(nodes2)

    filename2 = filename[:-5]+"_order2.mesh"
    mesh.Save(filename2)


def write_nas2mfem(filename,  reader, exclude_bdr=None, offset=None,
                   skip_unused_bdry=True, ho_thre=1e-20, skip_ho=False):

    geom_type = {'TETRA': 4,
                 'TRIA6': 2,
                 'TRIA3': 2,
                 'HEXA':  5,
                 'QUAD8': 3,
                 'QUAD4': 3, }
    '''                     
        SEGMENT = 1
        TRIANGLE = 2
        SQUARE = 3
        TETRAHEDRON = 4
        CUBE = 5
        '''

    if exclude_bdr is None:
        exclude_bdr = []
    if offset is None:
        offset = [0.0, 0.0, 0.0]
    if reader.dataset is None:
        reader.load()

    data = reader.dataset
    fid = open(filename, 'w')

    grid = data['GRIDS']
    elems = data['ELEMS']

    el_3d = ['TETRA', 'HEXA']
    if 'TETRA' in elems:
        el_2d = ['TRIA6', 'TRIA3']
    else:
        el_2d = ['QUAD8', 'QUAD4']

    n3d = np.sum([len(elems[x]) for x in el_3d if x in elems])

    if n3d > 0:
        unique_grids = list(np.unique(np.hstack([elems[name].flatten()
                                                 for name in el_3d if name in elems])))
    else:
        unique_grids = list(np.unique(np.hstack([elems[name].flatten()
                                                 for name in el_2d if name in elems])))
    nvtc = len(unique_grids)
    print('unique_grid (done)....' + str(nvtc))

    ndim = grid.shape[-1]
    nelem = 0
    nbdry = 0

    for k in elems:
        if k in el_3d:
            nelem = nelem + len(elems[k+'_ATTR'])
        if k in el_2d:
            tmp = [x for x in elems[k+'_ATTR'] if not x in exclude_bdr]
            nbdry = nbdry + len(tmp)

    fid.write('MFEM mesh v1.0\n')
    fid.write('\n')
    fid.write('dimension\n')
    fid.write(str(ndim) + '\n')
    fid.write('\n')
    fid.write('elements\n')
    fid.write(str(nelem) + '\n')

    rev_map = {unique_grids[x]: x for x in range(len(unique_grids))}
    reader.grid_mapping = (unique_grids, rev_map)

    for name in el_3d:
        if not name in elems:
            continue
        vidx = elems[name]
        attr = elems[name+'_ATTR']
        gtyp = geom_type[name]
        txts = [None]*len(attr)
        for i in range(len(attr)):
            txt = [str(attr[i]), str(gtyp)]
#                txt.extend([str(unique_grids.index(x)) for x in vidx[i]])
            txt.extend([str(rev_map[x]) for x in vidx[i]])
            txts[i] = ' '.join(txt)
        fid.write('\n'.join(txts))
    fid.write('\n')

    # count valid 2d elements
    # sometimes .nas contains a boundary element which are not used
    # in 3D mesh. By default we skip this

    bdry_checks = {}
    bdry_flags = {}
    for name in el_2d:
        if not name in elems:
            continue
        vidx = elems[name]
        ss = vidx.shape
        bdry_check = (np.in1d(vidx, unique_grids)).reshape(ss)
        bdry_flag = [len(x) == np.sum(x) for x in bdry_check]
        bdry_checks[name] = bdry_check
        bdry_flags[name] = bdry_flag

    n_validbdry = np.sum([np.sum(bdry_flags[x]) for x in bdry_flags])

    fid.write('boundary\n')
    fid.write(str(n_validbdry) + '\n')

    print("number of bdry in file:", nbdry)
    print("number of used bdry in file:", n_validbdry)
    for name in el_2d:
        if not name in elems:
            continue

        vidx = elems[name]

        attr = elems[name+'_ATTR']
        gtyp = geom_type[name]
        txts = [None]*n_validbdry

        bdry_flag = bdry_flags[name]
        k = 0
        for i in range(len(attr)):
            if attr[i] in exclude_bdr:
                continue
            if not bdry_flag[i]:
                continue
            txt = [str(attr[i]), str(gtyp)]
#                txt.extend([str(unique_grids.index(x)) for x in vidx[i]])
            txt.extend([str(rev_map[x]) for x in vidx[i]])
            txts[k] = ' '.join(txt)
            k = k + 1

        fid.write('\n'.join(txts))
    fid.write('\n')
    print("Writing vertices", nvtc)
    fid.write('vertices\n')
    fid.write(str(nvtc) + '\n')
    fid.write(str(ndim) + '\n')
    txts = [None]*nvtc
    for i in range(nvtc):
        txt = [str(x+offset[kk]) for kk, x in enumerate(grid[unique_grids[i]])]
        txts[i] = ' '.join(txt)
    fid.write('\n'.join(txts))
    print("Done")

    if reader.has_ho and not skip_ho:
        print("generating 2nd order mesh. this may take a while")
        write_ho_tet_mesh(filename, reader, ho_thre=ho_thre)
    fid.close()



import os
import numpy as np

from petram.model import Model
from petram.solver.solver_model import Solver
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('SetVar')
rprint = debug.regular_print('StdVar')

class SetVar(Solver):
    has_2nd_panel = False        
    def panel1_param(self):
        return [["physics model",   self.phys_model,  0, {},],
                ["Initial value setting",   self.init_setting,   0, {},],]

    def import_panel1_value(self, v):
        self.phys_model = str(v[0])        
        self.init_setting = str(v[1])
        
    def attribute_set(self, v):
        v['init_setting']   = ''
        super(SetVar, self).attribute_set(v)
        return v

    def get_panel1_value(self):
        return (self.phys_model, self.init_setting, )
    
    def get_init_setting(self):
        names = self.init_setting.split(',')
        names = [n.strip() for n in names if n.strip() != '']        
        return [self.root()['InitialValue'][n] for n in names]

    def get_matrix_weight(self, timestep_config):#, timestep_weight):
        return [0, 0, 0]

    def run(self, engine, is_first = True):

        self.init_only = True
        
        from petram.solver.solver_model import SolverInstance
        instance = SolverInstance(self, engine)
        
        phys_range = self.get_phys_range()                
        inits = self.get_init_setting()
        dprint1("Setting variable :" + self.name(), inits)        
        engine.run_apply_init(phys_range, inits=inits)

        phys_target = self.get_phys()        
        engine.save_sol_to_file(phys_target, 
                                skip_mesh = False,
                                mesh_only = False,
                                save_parmesh = self.save_parmesh)
        





import sys
model = proj.model1.mfem.param.eval("mfem_model")
for x in model.walk():
    mod = x.__class__.__module__
    cls = x.__class__.__name__
    newmod = '.'.join(['petram']+mod.split('.')[2:])
    mod = __import__(newmod)
    mod = sys.modules[newmod]
    newcls = getattr(mod, cls)
    x.__class__ = newcls
    



def eval_expr(expr, battrs, phys = None):
    '''
    expr = expression to evaluate
    battrs = list of boundary attribute
    '''
    from petram.model import Bdry
    from petram.utils import eval_expression
    m0 = model.param.getvar('mfem_model')
    if phys is None: 
        phys = m0['Phys'][m0['Phys'].keys()[0]]
    else:
        phys = m0['Phys'][phys]

    ret = {}
    engine.assign_sel_index(phys)
    for m in phys.walk():
       if isinstance(m, Bdry):
          for battr in battrs:
             #if battr in m._sel_index and not battr in ret:
              ns = m._global_ns
              ind_vars = m.get_independent_variables()
              ret[battr] = eval_expression(expr, mesh, battr, ind_vars, ns)
    return ret
ans(eval_expr(*args, **kwargs))



'''

Krylov Model. 
  Krylov solver
  This model is supposeed to work inside Iteratife solver chain.
  This model itsel does not define an iterative solver itself.

'''
from petram.solver.mumps_model import MUMPSPreconditioner
from petram.mfem_config import use_parallel
import numpy as np

from petram.debug import flush_stdout
from petram.namespace_mixin import NS_mixin
from .solver_model import LinearSolverModel, LinearSolver

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('KrylovModel')

if use_parallel:
    from petram.helper.mpi_recipes import *
    from mfem.common.parcsr_extra import *
    import mfem.par as mfem
    default_kind = 'hypre'

    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    smyid = '{:0>6d}'.format(myid)
    from mfem.common.mpi_debug import nicePrint

else:
    import mfem.ser as mfem
    default_kind = 'scipy'
    smyid = ''

choices_a = ['CG', 'GMRES', 'FGMRES', 'BiCGSTAB', 'MINRES']

single1_elp = [["log_level", -1, 400, {}],
               ["max  iter.", 200, 400, {}],
               ["rel. tol", 1e-7, 300, {}],
               ["abs. tol.", 1e-7, 300, {}], ]
single2_elp = [["log_level", -1, 400, {}],
               ["max  iter.", 200, 400, {}],
               ["rel. tol", 1e-7, 300, {}],
               ["abs. tol.", 1e-7, 300, {}],
               ["restart(kdim)", 50, 400, {}]]


class KrylovModel(LinearSolverModel, NS_mixin):
    hide_ns_menu = True
    has_2nd_panel = False
    accept_complex = False
    always_new_panel = False

    def __init__(self, *args, **kwargs):
        LinearSolverModel.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def init_solver(self):
        pass

    def panel1_param(self):
        return [[None, None, 34, [{'text': "Solver", 'choices': choices_a},
                                  {'elp': single1_elp},  # CG
                                  {'elp': single2_elp},  # GMRES
                                  {'elp': single2_elp},  # FGMRES
                                  {'elp': single1_elp},  # BiCGSTAB
                                  {'elp': single1_elp},  # MINRES
                                  ], ],
                [None, self.write_mat, 3, {"text": "write matrix"}],
                [None, self.assert_no_convergence, 3,
                    {"text": "check converegence"}], ]

    def get_panel1_value(self):
        # this will set _mat_weight
        from petram.solver.solver_model import SolveStep
        p = self.parent
        while not isinstance(p, SolveStep):
            p = p.parent
            if p is None:
                assert False, "Solver is not under SolveStep"

        single1 = [int(self.log_level), int(self.maxiter),
                   self.reltol, self.abstol]
        single2 = [int(self.log_level), int(self.maxiter),
                   self.reltol, self.abstol, int(self.kdim)]
        value = ([self.solver_type, single1, single2, single2,
                  single1, single1],
                 self.write_mat, self.assert_no_convergence,)

        return value

    def import_panel1_value(self, v):
        self.solver_type = str(v[0][0])
        idx = choices_a.index(self.solver_type)
        vv = v[0][idx + 1]
        self.log_level = int(vv[0])
        self.maxiter = int(vv[1])
        self.reltol = vv[2]
        self.abstol = vv[3]

        if len(vv) > 4:
            self.kdim = int(vv[4])

        self.write_mat = bool(v[1])
        self.assert_no_convergence = bool(v[2])

    def attribute_set(self, v):
        v = super(KrylovModel, self).attribute_set(v)
        v['solver_type'] = 'GMRES'
        v['log_level'] = 0
        v['maxiter'] = 200
        v['reltol'] = 1e-7
        v['abstol'] = 1e-7
        v['kdim'] = 50

        v['printit'] = 1
        v['write_mat'] = False
        v['assert_no_convergence'] = True
        return v

    def get_possible_child(self):
        from petram.solver.mumps_model import MUMPSMFEMSolverModel as MUMPS
        from petram.solver.block_smoother import DiagonalPreconditioner

        return KrylovModel, MUMPS, DiagonalPreconditioner

    def get_possible_child_menu(self):
        from petram.solver.mumps_model import MUMPSMFEMSolverModel as MUMPS
        from petram.solver.block_smoother import DiagonalPreconditioner

        choice = [("Preconditioner", KrylovSmoother),
                  ("", MUMPS),
                  ("!", DiagonalPreconditioner), ]
        return choice

    def verify_setting(self):
        return True, "", ""

    def real_to_complex(self, solall, M):
        if self.merge_real_imag:
            return self.real_to_complex_merged(solall, M)
        else:
            return self.real_to_complex_interleaved(solall, M)

    def real_to_complex_interleaved(self, solall, M):
        if use_parallel:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank

            offset = M.RowOffsets().ToList()
            of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o)))
                  for o in offset]
            if myid != 0:
                return

        else:
            offset = M.RowOffsets()
            of = offset.ToList()

        rows = M.NumRowBlocks()
        s = solall.shape
        nb = rows // 2
        i = 0
        pt = 0
        result = np.zeros((s[0] // 2, s[1]), dtype='complex')
        for j in range(nb):
            l = of[i + 1] - of[i]
            result[pt:pt + l, :] = (solall[of[i]:of[i + 1], :]
                                    + 1j * solall[of[i + 1]:of[i + 2], :])
            i = i + 2
            pt = pt + l

        return result

    def real_to_complex_merged(self, solall, M):
        if use_parallel:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank

            offset = M.RowOffsets().ToList()
            of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o)))
                  for o in offset]
            if myid != 0:
                return

        else:
            offset = M.RowOffsets()
            of = offset.ToList()

        rows = M.NumRowBlocks()
        s = solall.shape
        i = 0
        pt = 0
        result = np.zeros((s[0] // 2, s[1]), dtype='complex')
        for i in range(rows):
            l = of[i + 1] - of[i]
            w = int(l // 2)
            result[pt:pt + w, :] = (solall[of[i]:of[i] + w, :]
                                    + 1j * solall[(of[i] + w):of[i + 1], :])
            pt = pt + w
        return result

    def prepare_preconditioner(self, opr, engine):
        for x in self.iter_enabled():
            if isinstance(x, LinearSolverModel):
                return x.prepare_solver(opr, engine)


    def do_prepare_solver(self, opr, engine):
        cls = getattr(mfem, self.solver_type + 'Solver')
        args = (MPI.COMM_WORLD,) if use_parallel else ()

        solver = cls(*args)
        solver.SetAbsTol(self.abstol)
        solver.SetRelTol(self.reltol)
        solver.SetMaxIter(self.maxiter)
        solver.SetPrintLevel(self.log_level)
        if self.solver_type in ['GMRES', 'FGMRES']:
            solver.SetKDim(int(self.kdim))

        solver.iterative_mode = True
        solver.SetOperator(opr)
        if self.write_mat:
            self.write_matrix(opr)

        M = self.prepare_preconditioner(opr, engine)

        if M is not None:
            solver.SetPreconditioner(M)
            solver._prc = M
        return solver

    def write_matrix(self, A=None, b=None, x=None, suffix=smyid):
        def get_block(Op, i, j):
            try:
                return Op._linked_op[(i, j)]
            except KeyError:
                return None

        offset = A.RowOffsets()
        rows = A.NumRowBlocks()
        cols = A.NumColBlocks()
        if suffix != '':
            suffix = '.'+suffix

        if A is not None:
            for i in range(cols):
                for j in range(rows):
                    m = get_block(A, i, j)
                    if m is None:
                        continue
                    if isinstance(m, mfem.ComplexOperator):
                        m1 = m._real_operator
                        m2 = m._imag_operator
                        m1.Print('matrix_Re_' + str(i) + '_' + str(j))
                        m2.Print('matrix_Im_' + str(i) + '_' + str(j))
                    else:
                        m.Print('matrix_' + str(i) + '_' + str(j))
        if b is not None:
            for i, bb in enumerate(b):
                for j in range(rows):
                    v = bb.GetBlock(j)
                    v.Print('rhs_' + str(i) + '_' + str(j) + suffix)
        if x is not None:
            for j in range(rows):
                xx = x.GetBlock(j)
                xx.Print('x_' + str(i) + '_' + str(j) + suffix)

    def prepare_solver(self, opr, engine):
        solver = self.do_prepare_solver(opr, engine)
        solver.iterative_mode = True

        return solver

    @classmethod
    def fancy_menu_name(self):
        return 'KrylovSolver'

    @classmethod
    def fancy_tree_name(self):
        return 'Krylov'

    def get_info_str(self):
        return 'Solver'

    def does_linearsolver_choose_linearsystem_type(self):
        return False

    def supported_linear_system_type(self):
        return ["blk_interleave",
                "blk_merged_s",
                "blk_merged", ]


class KrylovSmoother(KrylovModel):
    @classmethod
    def fancy_menu_name(self):
        return 'KrylovSmoother'

    @classmethod
    def fancy_tree_name(self):
        return 'Krylov'

    def get_info_str(self):
        return 'Smoother'

    def attribute_set(self, v):
        v = KrylovModel.attribute_set(self, v)
        v['log_level'] = -1
        v['abstol'] = 0.0
        v['assert_no_convergence'] = False
        return v

    def prepare_solver(self, opr, engine):
        solver = self.do_prepare_solver(opr, engine)
        solver.iterative_mode = False
        return solver

    def prepare_solver_with_multtranspose(self, opr, engine):
        '''
        This is called from multi-level refinement
        '''
        class MyPySolver(mfem.PyIterativeSolver):
            def __init__(self, solver_type):
                args = (MPI.COMM_WORLD,) if use_parallel else ()
                cls = getattr(mfem, solver_type + 'Solver')
                self.s1 = cls(*args)
                self.s2 = cls(*args)
                mfem.PyIterativeSolver.__init__(self, *args)

            def SetAbsTol(self, tol):
                self.s1.SetAbsTol(tol)
                self.s2.SetAbsTol(tol)

            def SetRelTol(self, tol):
                self.s1.SetRelTol(tol)
                self.s2.SetRelTol(tol)

            def SetMaxIter(self, iter):
                self.s1.SetMaxIter(iter)
                self.s2.SetMaxIter(iter)

            def SetPrintLevel(self, lvl):
                self.s1.SetPrintLevel(lvl)
                self.s2.SetPrintLevel(lvl)

            def SetKDim(self, d):
                self.s1.SetKDim(d)
                self.s2.SetKDim(d)

            def SetOperator(self, opr):
                self.s1.SetOperator(opr)
                opr2 = mfem.TransposeOperator(opr)
                self.s2.SetOperator(opr2)
                self._oprs = (opr, opr2)

            def SetPreconditioner(self, opr):
                self.s1.SetPreconditioner(opr)
                #opr2 = mfem.TransposeOperator(opr)
                # self.s2.SetPreconditioner(opr2)
                self._prcs = (opr, )

            def Mult(self, x, y):
                self.s1.Mult(x, y)

            def MultTranspose(self, x, y):
                self.s1.Mult(x, y)

            @property
            def iterative_mode(self):
                return self.s1.iterative_mode

            @iterative_mode.setter
            def iterative_mode(self, v):
                self.s1.iterative_mode = v
                self.s2.iterative_mode = v

        args = (MPI.COMM_WORLD,) if use_parallel else ()
        solver = MyPySolver(self.solver_type)
        solver.SetAbsTol(self.abstol)
        solver.SetRelTol(self.reltol)
        solver.SetMaxIter(self.maxiter)
        solver.SetPrintLevel(self.log_level)
        if self.solver_type in ['GMRES', 'FGMRES']:
            solver.SetKDim(int(self.kdim))

        M = self.prepare_preconditioner(opr, engine)
        if M is not None:
            solver.SetPreconditioner(M)

        solver.iterative_mode = False
        solver.SetOperator(opr)
        return solver


class StationaryRefinementModel(KrylovModel):
    hide_ns_menu = True
    has_2nd_panel = False
    accept_complex = False
    always_new_panel = False

    def panel1_param(self):
        p1 = [["log_level", -1, 400, {}],
              ["max  iter.", 200, 400, {}],
              ["rel. tol", 1e-7, 300, {}],
              ["abs. tol.", 1e-7, 300, {}], ]
        return p1

    def get_panel1_value(self):
        # this will set _mat_weight
        from petram.solver.solver_model import SolveStep
        p = self.parent
        while not isinstance(p, SolveStep):
            p = p.parent
            if p is None:
                assert False, "Solver is not under SolveStep"

        single1 = [int(self.log_level), int(self.maxiter),
                   self.reltol, self.abstol]
        return single1

    def import_panel1_value(self, v):
        self.log_level = int(v[0])
        self.maxiter = int(v[1])
        self.reltol = v[2]
        self.abstol = v[3]

    def attribute_set(self, v):
        v = super(StationaryRefinementModel, self).attribute_set(v)
        v['log_level'] = 0
        v['maxiter'] = 200
        v['reltol'] = 1e-7
        v['abstol'] = 1e-7
        v['printit'] = 1
        v['write_mat'] = False
        v['assert_no_convergence'] = False
        return v

    def do_prepare_solver(self, opr, engine):
        cls = mfem.SLISolver
        args = (MPI.COMM_WORLD,) if use_parallel else ()

        solver = cls(*args)
        solver.SetAbsTol(self.abstol)
        solver.SetRelTol(self.reltol)
        solver.SetMaxIter(self.maxiter)
        solver.SetPrintLevel(self.log_level)

        solver.iterative_mode = True
        solver.SetOperator(opr)

        M = self.prepare_preconditioner(opr, engine)

        if M is not None:
            solver.SetPreconditioner(M)
            solver._prc = M
        return solver

    def prepare_solver(self, opr, engine):
        solver = self.do_prepare_solver(opr, engine)
        solver.iterative_mode = True

        return solver

    @classmethod
    def fancy_menu_name(self):
        return 'StationaryRefinement'

    @classmethod
    def fancy_tree_name(self):
        return 'Refinement'

    def get_info_str(self):
        return 'Refinement'

    def does_linearsolver_choose_linearsystem_type(self):
        return False

    def supported_linear_system_type(self):
        return ["blk_interleave",
                "blk_merged_s",
                "blk_merged", ]


class StationaryRefinementSmoother(StationaryRefinementModel):
    def get_info_str(self):
        return 'Smoother'

    def attribute_set(self, v):
        v = KrylovModel.attribute_set(self, v)
        v['log_level'] = -1
        v['abstol'] = 0.0
        v['assert_no_convergence'] = False
        return v

    def prepare_solver(self, opr, engine):
        solver = self.do_prepare_solver(opr, engine)
        solver.iterative_mode = False
        return solver
